##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Demonstration and test flowsheet for a dynamic flowsheet.

"""
from __future__ import division
from __future__ import print_function

# Import Python libraries
import logging

# Import Pyomo libraries
from pyomo.environ import ConcreteModel, SolverFactory, \
                            TransformationFactory, Var

# Import IDAES core
from idaes.core import FlowsheetBlockData, declare_process_block_class

# Import Property Modules
import reaction_properties_demo as properties_rxn

# Import Unit Model Modules
from idaes.models.cstr import CSTR

__author__ = "Andrew Lee"


# Set up logger
logger = logging.getLogger('idaes.flowsheet')


@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    """
    Create the flowsheet class.  Contains all the unit models and
    connections between them.  Also contains the main Pyomo model and
    solver.
    """

    def build(self):
        """
        Make the flowsheet object, fix some variables, and solve the problem
        """
        # Call UnitModel.build to setup dynamics
        super(_Flowsheet, self).build()

        # Add property packages to flowsheet library
        self.properties_rxn = properties_rxn.PropertyParameterBlock()

        # Create unit models
        self.Tank = CSTR(property_package=self.properties_rxn)

        # Add constraints to Tank
        self.Tank.height = Var(self.Tank.time,
                               initialize=0.68,
                               doc="Depth of fluid in tank")
        self.Tank.area = Var(initialize=1.0,
                             doc="Cross-sectional area of tank")

        @self.Tank.Constraint(self.Tank.time, doc="Tank geometry constraint")
        def geometry(b, t):
            return b.volume[t] == b.area*b.height[t]

        self.Tank.volume_flow = Var(self.Tank.time,
                                    initialize=154100.0,
                                    doc="Volumetric flow leaving tank")

        @self.Tank.Constraint(self.Tank.time, doc="Flow volume constraint")
        def volume_flow_calculation(b, t):
            return b.volume_flow[t] == (
                        b.holdup.properties_out[t].flow_mol *
                        b.holdup.properties_out[t].dens_mol_phase['Liq'])

        self.Tank.flow_coeff = Var(self.Tank.time,
                                   initialize=250000.0,
                                   doc="Tank outlet flow coefficient")

        @self.Tank.Constraint(self.Tank.time, doc="Outlet flow correlation")
        def outlet_flowrate(b, t):
            return b.volume_flow[t] == b.flow_coeff[t]*b.height[t]

        # Transform time domain
        discretizer = TransformationFactory('dae.finite_difference')
        discretizer.apply_to(self,
                             nfe=100,
                             wrt=self.time,
                             scheme='BACKWARD')

        # Call post_transform_build (needed for now)
        self.post_transform_build()

        # Add Connections

        # Expand connections
        TransformationFactory('network.expand_arcs').apply_to(self)


def build_model():
    """
    Construct flowsheet object
    """
    # Create a Concrete Model as the top level object
    m = ConcreteModel()

    # Creat flowsheet object and attach to Concrete Model
    m.fs_obj = Flowsheet(dynamic=True, time_set=[0, 1.0, 10000.0])

    return m


def setInputs(m):
    """
    Set inlet and operating conditions, and some initial conditions.
    """
    # ---------------------------------------------------------------------
    # Unit Tank
    m.fs_obj.Tank.inlet[:].vars["flow_mol_comp"]["a"].fix(1.0)
    m.fs_obj.Tank.inlet[:].vars["flow_mol_comp"]["b"].fix(2.0)
    m.fs_obj.Tank.inlet[:].vars["flow_mol_comp"]["c"].fix(0.1)
    m.fs_obj.Tank.inlet[:].vars["flow_mol_comp"]["d"].fix(0.0)
    m.fs_obj.Tank.inlet[:].vars["flow_mol_comp"]["e"].fix(0.0)
    m.fs_obj.Tank.inlet[:].vars["flow_mol_comp"]["f"].fix(0.0)
    m.fs_obj.Tank.inlet[:].vars["enth_mol"].fix(-238.71)
    m.fs_obj.Tank.inlet[:].vars["pressure"].fix(101325.0)

    m.fs_obj.Tank.flow_coeff.fix(250000.0)
    m.fs_obj.Tank.area.fix(1.0)
    m.fs_obj.Tank.heat.fix(0.0)

    # Initial Condition - Steady-State
    m.fs_obj.Tank.holdup.material_accumulation[0, "Liq", "a"].fix(0.0)
    m.fs_obj.Tank.holdup.material_accumulation[0, "Liq", "b"].fix(0.0)
    m.fs_obj.Tank.holdup.material_accumulation[0, "Liq", "c"].fix(0.0)
    m.fs_obj.Tank.holdup.material_accumulation[0, "Liq", "d"].fix(0.0)
    m.fs_obj.Tank.holdup.material_accumulation[0, "Liq", "e"].fix(0.0)
    m.fs_obj.Tank.holdup.material_accumulation[0, "Liq", "f"].fix(0.0)
    m.fs_obj.Tank.holdup.energy_accumulation[0, "Liq"].fix(0.0)

    # Give an initial value for volume to help initialization
    m.fs_obj.Tank.volume = 0.68


def initialize_model(m, solver, stee=False):
    """
    Set up and execute initializeation routine for model
    """
    # Initialize Units
    m.fs_obj.Tank.initialize(outlvl=1)

    # Solve flowsheet
    results = solver.solve(m.fs_obj, tee=stee)

    return results


def simulation1(m, solver, stee=False):
    # -------------------------------------------------------------------------
    # Create a step change in feed
    for t in m.fs_obj.time:
        if t > 5.0:
            # After time = 5.0, 50% increase in feed rate
            m.fs_obj.Tank.inlet[t].vars["flow_mol_comp"]["a"].fix(1.5)
            m.fs_obj.Tank.inlet[t].vars["flow_mol_comp"]["b"].fix(3.0)
            m.fs_obj.Tank.inlet[t].vars["flow_mol_comp"]["c"].fix(0.2)
            m.fs_obj.Tank.inlet[t].vars["flow_mol_comp"]["d"].fix(0.0)
            m.fs_obj.Tank.inlet[t].vars["flow_mol_comp"]["e"].fix(0.0)
            m.fs_obj.Tank.inlet[t].vars["flow_mol_comp"]["f"].fix(0.0)
            m.fs_obj.Tank.inlet[t].vars["enth_mol"].fix(-358.065)

    # -------------------------------------------------------------------------
    # Solve flowsheet
    results = solver.solve(m, tee=stee)

    return results


def print_summary(fs_obj):
    """
    Print some key results from the model.
    """
    print()
    print("Results")
    print()
    print("Tank")
    fs_obj.Tank.outlet.display()
    fs_obj.Tank.height.display()
    fs_obj.Tank.volume_flow.display()


def main():
    """
    Make the flowsheet object, fix some variables, and solve the problem
    """
    # Create a Concrete Model as the top level object
    m = build_model()

    # Fix variables
    setInputs(m)

    # Create a solver
    stee = True
    opt = SolverFactory('ipopt')
    opt.options = {'tol': 1e-6,
                   'mu_init': 1e-8,
                   'bound_push': 1e-8}

    m.fs_obj.model_check()

    logger.info("Begin Flowsheet initialisation.")
    initialize_model(m, opt, stee)
    logger.info("Flowsheet initialisation finished.")

    results = simulation1(m, opt, stee)

    # Print results
    print(results)
    print_summary(m.fs_obj)


if __name__ == "__main__":
    main()
