##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Demonstration and test flowsheet for basic unit models.

"""
from __future__ import division
from __future__ import print_function

# Import Python libraries
import logging

# Import Pyomo libraries
from pyomo.environ import ConcreteModel, SolverFactory, \
            TransformationFactory, value

# Import IDAES core
from idaes.core import FlowsheetBlockData, declare_process_block_class

# Import Property Modules
import steam_properties_demo as properties

# Import Unit Model Modules
from idaes.models.pressure_changer \
        import PressureChanger as Compressor
from idaes.models.heat_exchanger import HeatExchanger as HX

__author__ = "Andrew Lee"


# Set up logger
logger = logging.getLogger('idaes.flowsheet')


@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    """
    Create the flowsheet class.  Contains all the unit models and
    connections between them.  Also contains the main Pyomo model and
    solver.
    """

    def build(self):
        """
        Make the flowsheet object, fix some variables, and solve the problem
        """
        # Call UnitModel.build to setup dynamics
        super(_Flowsheet, self).build()

        # Add property packages to flowsheet library
        self.steam_props = properties.PropertyParameterBlock()
        self.config.default_property_package = self.steam_props

        # Create unit models
        self.C01 = Compressor(thermodynamic_assumption="isentropic",
                              property_package=self.steam_props,
                              inlet_list=["1", "2"])

        self.HX02 = HX(flow_type="counter-current",
                       side_1_property_package=self.steam_props,
                       side_1_outlet_list=["1", "2"])

        # Transform time domain
        discretizer = TransformationFactory('dae.finite_difference')
        discretizer.apply_to(self,
                             nfe=2,
                             wrt=self.time,
                             scheme='BACKWARD')

        # Call post_transform_build (needed for now)
        self.post_transform_build()

        # Add Connections
        @self.Arc(self.time, doc="Connection between C01 and HX02",
            directed=True)
        def connection_1(b, t):
            return (b.C01.outlet[t], b.HX02.side_1_inlet[t])

        # Expand connections
        TransformationFactory('network.expand_arcs').apply_to(self)


def build_model():
    """
    Construct flowsheet object
    """
    # Create a Concrete Model as the top level object
    m = ConcreteModel()

    # Creat flowsheet object and attach to Concrete Model
    m.fs_obj = Flowsheet(dynamic=True, time_set=[0, 1.0])

    return m


def setInputs(m):
    """
    Set inlet and operating conditions.
    """

    # ---------------------------------------------------------------------
    # Unit C01
    m.fs_obj.C01.inlet[:, "1"].vars["flow_mol_comp"]["H2O"].fix(27.5e3)
    m.fs_obj.C01.inlet[:, "1"].vars["temperature"].fix(866.5)
    m.fs_obj.C01.inlet[:, "1"].vars["pressure"].fix(2.97e7)

    m.fs_obj.C01.inlet[:, "2"].vars["flow_mol_comp"]["H2O"].fix(27.5e3)
    m.fs_obj.C01.inlet[:, "2"].vars["temperature"].fix(866.5)
    m.fs_obj.C01.inlet[:, "2"].vars["pressure"].fix(2.97e7)

    m.fs_obj.C01.volume.fix(1.0)
    m.fs_obj.C01.deltaP.fix(-1e7)
    m.fs_obj.C01.efficiency_isentropic.fix(0.83)

    m.fs_obj.C01.holdup.material_accumulation[0, "Vap", "H2O"].fix(0.0)
    m.fs_obj.C01.holdup.energy_accumulation[0, "Vap"].fix(0.0)

    # ---------------------------------------------------------------------
    # Unit HX02
    m.fs_obj.HX02.side_2_inlet[:].vars["flow_mol_comp"]["H2O"].fix(27.5e3)
    m.fs_obj.HX02.side_2_inlet[:].vars["temperature"].fix(500.0)
    m.fs_obj.HX02.side_2_inlet[:].vars["pressure"].fix(2.97e7)

    m.fs_obj.HX02.volume_side_1.fix(1.0)
    m.fs_obj.HX02.volume_side_2.fix(1.0)
    m.fs_obj.HX02.heat_transfer_coefficient.fix(5.0)
    m.fs_obj.HX02.heat_transfer_area.fix(100000.0)

    m.fs_obj.HX02.side_1.outlet_splitter.split_fraction[:, "1"].fix(0.4)

    # Initial Condition - Steady-State
    m.fs_obj.HX02.side_1.material_accumulation[0, "Vap", "H2O"].fix(0.0)
    m.fs_obj.HX02.side_1.energy_accumulation[0, "Vap"].fix(0.0)
    m.fs_obj.HX02.side_2.material_accumulation[0, "Vap", "H2O"].fix(0.0)
    m.fs_obj.HX02.side_2.energy_accumulation[0, "Vap"].fix(0.0)


def initialize_model(m, solver, stee=False):
    """
    Set up and execute initializeation routine for model
    """
    # Initialize Units
    m.fs_obj.C01.initialize(state_args={"flow_mol_comp": {"H2O": 27.5e3},
                                        "pressure": 2.97e7,
                                        "temperature": 866.5},
                            outlvl=1)

    m.fs_obj.HX02.side_1_inlet[0].vars["flow_mol_comp"]["H2O"].value = \
        value(m.fs_obj.C01.outlet[0].vars["flow_mol_comp"]["H2O"])
    m.fs_obj.HX02.side_1_inlet[0].vars["pressure"].value = \
        value(m.fs_obj.C01.outlet[0].vars["pressure"])
    m.fs_obj.HX02.side_1_inlet[0].vars["temperature"].value = \
        value(m.fs_obj.C01.outlet[0].vars["temperature"])

    m.fs_obj.HX02.initialize(outlvl=1)

    # Solve flowsheet
    results = solver.solve(m.fs_obj, tee=stee)

    return results


def simulation1(m, solver, stee=False):
    # Solve flowsheet
    results = solver.solve(m, tee=stee)

    return results


def print_summary(fs_obj):
    """
    Print some key results from the model.
    """
    print()
    print("Results")
    print()
    print("C01")
    fs_obj.C01.inlet.display()
    fs_obj.C01.outlet.display()
    fs_obj.C01.work_mechanical.display()
    print()
    print("HX02")
    fs_obj.HX02.side_1_inlet.display()
    fs_obj.HX02.side_2_inlet.display()
    fs_obj.HX02.side_1_outlet.display()
    fs_obj.HX02.side_2_outlet.display()
    fs_obj.HX02.heat.display()
    fs_obj.HX02.side_2.volume.display()


def main():
    """
    Make the flowsheet object, fix some variables, and solve the problem
    """
    # Create a Concrete Model as the top level object
    m = build_model()

    # Fix variables
    setInputs(m)

    # Create a solver
    stee = True
    opt = SolverFactory('ipopt')
    opt.options = {'tol': 1e-6,
                   'mu_init': 1e-8,
                   'bound_push': 1e-8}

    m.fs_obj.model_check()

    # Solve flowsheet
    logger.info("Begin Flowsheet initialisation.")
    initialize_model(m, opt, stee)
    logger.info("Flowsheet initialisation finished.")

    # Run simulation
    results = simulation1(m, opt, stee)

    # Print results
    print(results)
    print_summary(m.fs_obj)


if __name__ == "__main__":
    main()
