##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Demonstration property package for the Gibbs rector unit model
"""

# Chages the divide behavior to not do integer division
from __future__ import division

# Import Python libraries
import logging

# Import Pyomo libraries
from pyomo.environ import Constraint, cosh, Expression, log, Param, \
                          PositiveReals, Reals, Set, sinh, \
                          tanh, value, Var
from pyomo.opt import SolverFactory, TerminationCondition

# Import IDAES cores
from idaes.core import declare_process_block_class, PropertyBlockBase, \
                        PropertyBlockDataBase, PropertyParameterBase
from idaes.core.util.misc import add_object_ref, solve_indexed_blocks

# Some more inforation about this module
__author__ = "Andrew Lee"


# Set up logger
logger = logging.getLogger('idaes.unit_model.properties')


@declare_process_block_class("PropertyParameterBlock")
class PropertyParameterData(PropertyParameterBase):
    """
    Property Parameter Block Class

    Contains parameters and indexing sets associated with properties for
    superheated steam.

    """

    def build(self):
        '''
        Callable method for Block construction.
        '''
        super(PropertyParameterData, self).build()

        self._make_params()

    def _make_params(self):
        '''
        This section is for parameters needed for the property models.
        '''
        # List of valid phases in property package
        self.phase_list = Set(initialize=['Vap'])

        # Component list - a list of component identifiers
        self.component_list = Set(initialize=['H2', 'N2', 'O2', 'CH4',
                                              'CO', 'CO2', 'H2O', 'NH3'])

        # List of components in each phase (optional)
        self.phase_comp = {"Vap": self.component_list}

        # List of all chemical elements that constitute the chemical species
        self.element_list = Set(initialize=['H', 'N', 'O', 'C'])

        # Elemental composition of all species
        self.element_comp = {'H2': {'H': 2, 'N': 0, 'O': 0, 'C': 0},
                             'N2': {'H': 0, 'N': 2, 'O': 0, 'C': 0},
                             'O2': {'H': 0, 'N': 0, 'O': 2, 'C': 0},
                             'CH4': {'H': 4, 'N': 0, 'O': 0, 'C': 1},
                             'CO': {'H': 0, 'N': 0, 'O': 1, 'C': 1},
                             'CO2': {'H': 0, 'N': 0, 'O': 2, 'C': 1},
                             'H2O': {'H': 2, 'N': 0, 'O': 1, 'C': 0},
                             'NH3': {'H': 3, 'N': 1, 'O': 0, 'C': 0}}

        # DIPPR ideal gas heat capacity correlation parameters
        self.cpigdip = {'H2': {1: 27.617, 2: 9.5600, 3: 2466.0,
                               4: 3.76000, 5: 567.60},
                        'N2': {1: 29.105, 2: 8.6149, 3: 1701.6,
                               4: 0.10347, 5: 909.79},
                        'O2': {1: 29.103, 2: 10.040, 3: 2526.5,
                               4: 9.35600, 5: 1153.8},
                        'CH4': {1: 33.298, 2: 79.933, 3: 2086.9,
                                4: 41.6020, 5: 991.96},
                        'CO': {1: 29.108, 2: 8.7730, 3: 3085.1,
                               4: 8.45530, 5: 1538.2},
                        'CO2': {1: 29.370, 2: 34.540, 3: 1428.0,
                                4: 26.4000, 5: 588.00},
                        'H2O': {1: 33.363, 2: 26.790, 3: 2610.5,
                                4: 8.89600, 5: 1169.0},
                        'NH3': {1: 33.427, 2: 48.980, 3: 2036.0,
                                4: 22.5600, 5: 882.00}}

        # Gas Constant
        self.gas_const = Param(within=PositiveReals,
                               mutable=False,
                               default=8.314,
                               doc='Gas Constant [J/mol.K]')

        # Mixture heat capacity
        self.cp_mol = Param(within=PositiveReals,
                            mutable=True,
                            default=29.07,
                            doc="Molar heat capacity [J/mol.K]")

        # Thermodynamic reference state
        self.pressure_ref = Param(within=PositiveReals,
                                  mutable=True,
                                  default=101325.0,
                                  doc='Reference pressure [Pa]')
        self.temperature_ref = Param(within=PositiveReals,
                                     mutable=True,
                                     default=303.15,
                                     doc='Reference temperature [K]')

    def get_supported_properties(self):
        return {'flow_mol_comp': {'method': None, 'units': 'mol/s'},
                'pressure': {'method': None, 'units': 'Pa'},
                'temperature': {'method': None, 'units': 'K'},
                'mole_frac': {'method': None, 'units': None},
                'cp_mol': {'method': None, 'units': 'J/mol.K'},
                'cp_mol_vap_pc': {'method': '_cp_mol_vap_pc',
                                  'units': 'J/mol.K'},
                'dens_mol_phase': {'method': '_dens_mol_phase',
                                   'units': 'mol/m^3'},
                'enth_mol': {'method': '_enth_mol', 'units': 'J/mol'},
                'enth_mol_vap_pc': {'method': '_enth_mol_vap_pc',
                                    'units': 'J/mol'},
                'gibbs_pc': {'method': '_gibbs_pc', 'units': 'J/mol'}}

    def get_package_units(self):
        return {'time': 's',
                'length': 'm',
                'mass': 'g',
                'amount': 'mol',
                'temperature': 'K',
                'energy': 'J',
                'holdup': 'mol'}


class _PropertyBlock(PropertyBlockBase):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """
    def initialize(blk, flow_mol_comp=None, temperature=None, pressure=None,
                   hold_state=False, outlvl=0,
                   solver='ipopt', optarg={'tol': 1e-8}):
        '''
        Initialisation routine for property package.

        Keyword Arguments:
            flow_mol_component : value at which to initialize component flows
                             (default=None)
            pressure : value at which to initialize pressure (default=None)
            temperature : value at which to initialize temperature
                          (default=None)
            outlvl : sets output level of initialisation routine

                     * 0 = no output (default)
                     * 1 = return solver state for each step in routine
                     * 2 = include solver output infomation (tee=True)

            optarg : solver options dictionary object (default=None)
            solver : str indicating whcih solver to use during
                     initialization (default = 'ipopt')
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         - True - states varaibles are not unfixed, and
                                 a dict of returned containing flags for
                                 which states were fixed during
                                 initialization.
                        - False - state variables are unfixed after
                                 initialization by calling the
                                 relase_state method

        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        '''
        # Fix state variables if not already fixed
        Fcflag = {}
        Pflag = {}
        Tflag = {}

        for k in blk.keys():
            for j in blk[k].component_list:
                if blk[k].flow_mol_comp[j].fixed is True:
                    Fcflag[k, j] = True
                else:
                    Fcflag[k, j] = False
                    if flow_mol_comp is None:
                        blk[k].flow_mol_comp[j].fix(1.0)
                    else:
                        blk[k].flow_mol_comp[j].fix(flow_mol_comp[j])

            if blk[k].pressure.fixed is True:
                Pflag[k] = True
            else:
                Pflag[k] = False
                if pressure is None:
                    blk[k].pressure.fix(101325.0)
                else:
                    blk[k].pressure.fix(pressure)

            if blk[k].temperature.fixed is True:
                Tflag[k] = True
            else:
                Tflag[k] = False
                if temperature is None:
                    blk[k].temperature.fix(300.0)
                else:
                    blk[k].temperature.fix(temperature)

            for j in blk[k].component_list:
                blk[k].mole_frac[j] = (value(blk[k].flow_mol_comp[j]) /
                                       sum(value(blk[k].flow_mol_comp[i])
                                           for i in blk[k].component_list))

        # Set solver options
        if outlvl > 1:
            stee = True
        else:
            stee = False

        opt = SolverFactory(solver)
        opt.options = optarg

        # ---------------------------------------------------------------------
        # Initialise values
        for k in blk.keys():
            for j in blk[k].component_list:
                blk[k].cp_mol_vap_pc[j] = (
                        blk[k].cpigdip[j][1] + blk[k].cpigdip[j][2] *
                        ((blk[k].cpigdip[j][3]/value(blk[k].temperature)) /
                         sinh(blk[k].cpigdip[j][3] /
                              value(blk[k].temperature)))**2 +
                        blk[k].cpigdip[j][4] *
                        ((blk[k].cpigdip[j][5]/value(blk[k].temperature)) /
                         cosh(blk[k].cpigdip[j][5] /
                              value(blk[k].temperature)))**2)

                blk[k].enth_mol_vap_pc[j] = (
                        (blk[k].cpigdip[j][1]*value(blk[k].temperature) +
                         blk[k].cpigdip[j][2]*blk[k].cpigdip[j][3] /
                         tanh(blk[k].cpigdip[j][3]/value(blk[k].temperature)) -
                         blk[k].cpigdip[j][4]*blk[k].cpigdip[j][5] *
                         tanh(blk[k].cpigdip[j][5] /
                              value(blk[k].temperature))) -
                        (blk[k].cpigdip[j][1]*value(blk[k].temperature_ref()) +
                         blk[k].cpigdip[j][2]*blk[k].cpigdip[j][3] /
                         tanh(blk[k].cpigdip[j][3] /
                              value(blk[k].temperature_ref())) -
                         blk[k].cpigdip[j][4]*blk[k].cpigdip[j][5] *
                         tanh(blk[k].cpigdip[j][5] /
                              value(blk[k].temperature_ref()))))

                blk[k].gibbs_pc[j] = (
                    value(blk[k].cp_mol_vap_pc[j]) *
                    (value(blk[k].temperature) -
                     value(blk[k].temperature_ref())) -
                    value(blk[k].temperature)*value(blk[k].cp_mol_vap_pc[j]) *
                    log(value(blk[k].temperature) /
                        value(blk[k].temperature_ref())) +
                    value(blk[k].temperature)*value(blk[k].gas_const) *
                    log(value(blk[k].pressure) /
                        value(blk[k].pressure_ref())))

            blk[k].h_mix = sum(value(blk[k].mole_frac[j]) *
                               value(blk[k].enth_mol_vap_pc[j])
                               for j in blk[k].component_list)

        results = solve_indexed_blocks(opt, blk, tee=stee)

        if outlvl > 0:
            if results.solver.termination_condition \
                    == TerminationCondition.optimal:
                logger.info('{} Initialisation Step 1 Complete.'
                            .format(blk.name))
            else:
                logger.warning('{} Initialisation Step 1 Failed.'
                               .format(blk.name))

        # ---------------------------------------------------------------------
        # If input block, return flags, else release state
        flags = {"Fcflag": Fcflag, "Pflag": Pflag,
                 "Tflag": Tflag}

        if outlvl > 0:
            if outlvl > 0:
                logger.info('{} Initialisation Complete.'.format(blk.name))

        if hold_state is True:
            return flags
        else:
            blk.release_state(flags)

    def release_state(blk, flags, outlvl=0):
        '''
        Method to relase state variables fixed during initialisation.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of of logging
        '''
        # Unfix state variables
        for k in blk.keys():
            for j in blk[k].component_list:
                if flags['Fcflag'][k, j] is False:
                    blk[k].flow_mol_comp[j].unfix()
            if flags['Pflag'][k] is False:
                blk[k].pressure.unfix()
            if flags['Tflag'][k] is False:
                blk[k].temperature.unfix()

        if outlvl > 0:
            if outlvl > 0:
                logger.info('{} State Released.'.format(blk.name))


@declare_process_block_class("PropertyBlock",
                             block_class=_PropertyBlock)
class PropertyBlockData(PropertyBlockDataBase):
    """
    This is an example of a property package for calcuating the Gibbs energy
    minimisation.
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super(PropertyBlockData, self).build()

        self._make_params()
        self._make_vars()
        self._make_constraints()
        self._make_balance_terms()

    def _make_params(self):
        # Create package parameters
        ''' This section is for parameters needed for the property models.'''

        # List of valid phases in property package
        add_object_ref(self, "phase_list", self.config.parameters.phase_list)

        # Component list - a list of component identifiers
        add_object_ref(self, "component_list",
                       self.config.parameters.component_list)

        # List of all chemical elements that constitute the chemical species
        add_object_ref(self, "element_list",
                       self.config.parameters.element_list)

        # Elemental composition of all species
        add_object_ref(self, "element_comp",
                       self.config.parameters.element_comp)

        # DIPPR ideal gas heat capacity correlation parameters
        add_object_ref(self, "cpigdip",
                       self.config.parameters.cpigdip)

        # Gas constant
        add_object_ref(self, "gas_const",
                       self.config.parameters.gas_const)

        # Mixture heat capacity
        add_object_ref(self, "cp_mol", self.config.parameters.cp_mol)

        # Thermodynamic reference state
        add_object_ref(self, "pressure_ref",
                       self.config.parameters.pressure_ref)
        add_object_ref(self, "temperature_ref",
                       self.config.parameters.temperature_ref)

    def _make_vars(self):
        # Create state variables
        ''' This section contains the state variables to be used to calucate
            the properties, and needs to be linked to the state properties in
            the unit model.'''

        self.flow_mol_comp = Var(self.component_list,
                                 initialize=1.0,
                                 doc='Component molar flowrate [mol/s]')
        self.pressure = Var(domain=Reals,
                            initialize=101325.0,
                            bounds=(1e3, 1e6),
                            doc='State pressure [Pa]')
        self.temperature = Var(domain=Reals,
                               initialize=303.15,
                               bounds=(273.15, 500.0),
                               doc='State temperature [K]')
        self.mole_frac = Var(self.component_list,
                             domain=Reals,
                             initialize=0.0,
                             doc='State component mole fractions [-]')

    def _make_constraints(self):
        # Create standard constraints
        ''' This section creates the necessary constraints for calculating
            the standard property values.
            All calcuations assume ideal gas behaviour
        '''
        # Mole fractions
        def mole_frac_constraint(b, j):
            return b.flow_mol_comp[j] == (
                       b.mole_frac[j] *
                       sum(b.flow_mol_comp[k] for k in b.component_list))
        self.mole_frac_constraint = Constraint(self.component_list,
                                               rule=mole_frac_constraint)

    def _dens_mol_phase(self):
        # Molar density
        self.dens_mol_phase = Var(self.phase_list,
                                  doc="Molar density")

        def ideal_gas(b, p):
            return (b.dens_mol_phase[p]*b.gas_const*b.temperature ==
                    b.pressure)
        try:
            # Try to build constraint
            self.ideal_gas = Constraint(self.phase_list, rule=ideal_gas)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.dens_mol_phase)
            self.del_component(self.ideal_gas)
            raise

    def _cp_mol_vap_pc(self):
        # Pure component vapour heat capacities
        self.cp_mol_vap_pc = Var(self.component_list,
                                 domain=Reals,
                                 initialize=1.0,
                                 doc="Pure component vapour heat capacities "
                                 "[J/mol.K]")

        def pure_component_cp_mol(b, j):
            return b.cp_mol_vap_pc[j] == (
                                      b.cpigdip[j][1] +
                                      b.cpigdip[j][2] *
                                      ((b.cpigdip[j][3]/b.temperature) /
                                       sinh(b.cpigdip[j][3] /
                                            b.temperature))**2 +
                                      b.cpigdip[j][4] *
                                      ((b.cpigdip[j][5]/b.temperature) /
                                       cosh(b.cpigdip[j][5] /
                                            b.temperature))**2)
        try:
            # Try to build constraint
            self.pure_component_cp_mol = Constraint(
                                            self.component_list,
                                            rule=pure_component_cp_mol)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.cp_mol_vap_pc)
            self.del_component(self.pure_component_cp_mol)
            raise

    def _gibbs_pc(self):
        # Pure component Gibbs free energy
        self.gibbs_pc = Var(self.component_list,
                            domain=Reals,
                            initialize=0.0,
                            doc="Pure component Gibbs free energy [J/mol]")

        # Assume constant cp_mol for simplicity and vapour phase only
        def gibbs_energy_equation(b, j):
            return b.gibbs_pc[j] == (b.cp_mol_vap_pc[j]*(b.temperature -
                                     b.temperature_ref()) -
                                     b.temperature*b.cp_mol_vap_pc[j] *
                                     log(b.temperature/b.temperature_ref()) +
                                     b.temperature*b.gas_const *
                                     log(b.pressure/b.pressure_ref()))
        try:
            # Try to build constraint
            self.gibbs_energy_equation = Constraint(self.component_list,
                                                    rule=gibbs_energy_equation)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.gibbs_pc)
            self.del_component(self.gibbs_energy_equation)
            raise

    def _enth_mol(self):
        # Mixture molar enthalpy
        self.enth_mol = Var(domain=Reals,
                            initialize=0.0,
                            doc='Mixture specific entahlpy [J/mol]')
        try:
            # Try to build constraint
            self.mixture_enthalpy = Constraint(expr=(
                        self.enth_mol == sum(self.mole_frac[j] *
                                             self.enth_mol_vap_pc[j]
                                             for j in self.component_list)))
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.enth_mol)
            self.del_component(self.mixture_enthalpy)
            raise

    def _enth_mol_vap_pc(self):
        # Pure component vapour enthalpies
        self.enth_mol_vap_pc = Var(
                self.component_list,
                domain=Reals,
                initialize=1.0,
                doc="Pure component vapour enthalpies [J/mol]")

        def pure_comp_enthalpy(b, j):
            return b.enth_mol_vap_pc[j] == (
                        (b.cpigdip[j][1]*b.temperature +
                         b.cpigdip[j][2]*b.cpigdip[j][3] /
                         tanh(b.cpigdip[j][3]/b.temperature) -
                         b.cpigdip[j][4]*b.cpigdip[j][5] *
                         tanh(b.cpigdip[j][5]/b.temperature)) -
                        (b.cpigdip[j][1]*b.temperature_ref() +
                         b.cpigdip[j][2]*b.cpigdip[j][3] /
                         tanh(b.cpigdip[j][3] /
                              b.temperature_ref()) -
                         b.cpigdip[j][4]*b.cpigdip[j][5] *
                         tanh(b.cpigdip[j][5] /
                              b.temperature_ref())))
        try:
            # Try to build constraint
            self.pure_comp_enthalpy = Constraint(self.component_list,
                                                 rule=pure_comp_enthalpy)
        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.enth_mol_vap_pc)
            self.del_component(self.pure_comp_enthalpy)
            raise

    def _make_balance_terms(self):
        def material_terms(b, i, j):
            return b.flow_mol_comp[j]
        self.material_balance_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_terms)

        def energy_terms(b, i):
            return sum(b.flow_mol_comp[j] for j in b.component_list)*b.enth_mol
        self.energy_balance_term = Expression(self.phase_list,
                                              rule=energy_terms)

        def material_density_terms(b, p, j):
            return b.dens_mol_phase[p]*b.mole_frac[j]
        self.material_density_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_density_terms)

        def energy_density_terms(b, p):
            return b.dens_mol_phase[p]*b.enth_mol
        self.energy_density_term = Expression(self.phase_list,
                                              rule=energy_density_terms)

    def declare_port_members(b):
        members = {"flow_mol_comp": b.flow_mol_comp,
                   "temperature": b.temperature,
                   "pressure": b.pressure}
        return members

    def model_check(blk):
        """
        Model checks for property block
        """
        # Check temperature bounds
        if value(blk.temperature) < blk.temperature.lb:
            logger.error('{} Temperature set below lower bound.'
                         .format(blk.name))
        if value(blk.temperature) > blk.temperature.ub:
            logger.error('{} Temperature set above upper bound.'
                         .format(blk.name))

        # Check pressure bounds
        if value(blk.pressure) < blk.pressure.lb:
            logger.error('{} Pressure set below lower bound.'.format(blk.name))
        if value(blk.pressure) > blk.pressure.ub:
            logger.error('{} Pressure set above upper bound.'.format(blk.name))
