##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Demonstration and test flowsheet for a dynamic flowsheet.

"""
from __future__ import division
from __future__ import print_function

# Import Pyomo libraries
from pyomo.environ import ConcreteModel, minimize, Objective, SolverFactory

# Import IDAES core
from idaes.core import FlowsheetBlock, Stream

# Import properties
import tutorial_1_properties as properties_rxn

# Import Unit Model Modules
from idaes.models import Feed, CSTR, Product


def main():
    """
    Make the flowsheet object, fix some variables, and solve the problem
    """
    # Create a Concrete Model as the top level object
    m = ConcreteModel()

    # Add a flowsheet object to the model
    m.fs = FlowsheetBlock(dynamic=False)

    # Add property packages to flowsheet library
    m.fs.properties_rxn = properties_rxn.PropertyParameterBlock()

    # Create unit models
    m.fs.Feed = Feed(property_package=m.fs.properties_rxn)
    m.fs.Tank1 = CSTR(property_package=m.fs.properties_rxn)
    m.fs.Tank2 = CSTR(property_package=m.fs.properties_rxn)
    m.fs.Product = Product(property_package=m.fs.properties_rxn)

    # Post transform build
    m.fs.post_transform_build()

    # Make Streams
    m.fs.stream_1 = Stream(source=m.fs.Feed.outlet,
                           destination=m.fs.Tank1.inlet)
    m.fs.stream_2 = Stream(source=m.fs.Tank1.outlet,
                           destination=m.fs.Tank2.inlet)
    m.fs.stream_3 = Stream(source=m.fs.Tank2.outlet,
                           destination=m.fs.Product.inlet)

    # Set inlet and operating conditions, and some initial conditions.
    m.fs.Feed.fix('flow_mol_comp', comp='a', value=1.0)
    m.fs.Feed.fix('flow_mol_comp', comp='b', value=2.0)
    m.fs.Feed.fix('flow_mol_comp', comp='c', value=0.1)
    m.fs.Feed.fix('flow_mol_comp', comp='d', value=0.0)
    m.fs.Feed.fix('flow_mol_comp', comp='e', value=0.0)
    m.fs.Feed.fix('flow_mol_comp', comp='f', value=0.0)
    m.fs.Feed.fix('temperature', value=303.15)
    m.fs.Feed.fix('pressure', value=101325)

    m.fs.Tank1.volume.fix(10.0)
    m.fs.Tank1.heat.fix(0.0)

    m.fs.Tank2.volume.fix(10.0)
    m.fs.Tank2.heat.fix(0.0)

    # Initialize Units
    m.fs.Feed.initialize()
    m.fs.Tank1.initialize(state_args={
            "flow_mol_comp": {
                "a": m.fs.Feed.outlet[0].vars["flow_mol_comp"]["a"].value,
                "b": m.fs.Feed.outlet[0].vars["flow_mol_comp"]["b"].value,
                "c": m.fs.Feed.outlet[0].vars["flow_mol_comp"]["c"].value,
                "d": m.fs.Feed.outlet[0].vars["flow_mol_comp"]["d"].value,
                "e": m.fs.Feed.outlet[0].vars["flow_mol_comp"]["e"].value,
                "f": m.fs.Feed.outlet[0].vars["flow_mol_comp"]["f"].value},
            "pressure": m.fs.Feed.outlet[0].vars["pressure"].value,
            "temperature": m.fs.Feed.outlet[0].vars["temperature"].value})
    m.fs.Tank2.initialize(state_args={
            "flow_mol_comp": {
                "a": m.fs.Tank1.outlet[0].vars["flow_mol_comp"]["a"].value,
                "b": m.fs.Tank1.outlet[0].vars["flow_mol_comp"]["b"].value,
                "c": m.fs.Tank1.outlet[0].vars["flow_mol_comp"]["c"].value,
                "d": m.fs.Tank1.outlet[0].vars["flow_mol_comp"]["d"].value,
                "e": m.fs.Tank1.outlet[0].vars["flow_mol_comp"]["e"].value,
                "f": m.fs.Tank1.outlet[0].vars["flow_mol_comp"]["f"].value},
            "pressure": m.fs.Tank1.outlet[0].vars["pressure"].value,
            "temperature": m.fs.Tank1.outlet[0].vars["temperature"].value})

    # =========================================================================
    # Adding Optimization
    m.fs.obj = Objective(expr=m.fs.Tank2.outlet[0].vars["flow_mol_comp"]["f"],
                         sense=minimize)

    # Set bounds for variables
    m.fs.Tank1.outlet[0].vars["temperature"].setlb(300)
    m.fs.Tank2.outlet[0].vars["temperature"].setlb(300)
    m.fs.Tank1.outlet[0].vars["temperature"].setub(400)
    m.fs.Tank2.outlet[0].vars["temperature"].setub(450)

    # Unfix heat inputs
    m.fs.Tank1.heat.unfix()
    m.fs.Tank2.heat.unfix()
    # =========================================================================

    # Create a solver
    solver = SolverFactory('ipopt')
    results = solver.solve(m, tee=True)

    # Print results
    print(results)
    print()
    print("Results")
    print()
    m.fs.Product.display()
    m.fs.Tank1.heat.display()
    m.fs.Tank2.heat.display()

    # For testing purposes
    return(m, results)


if __name__ == "__main__":
    main()
