##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Demonstration property package for water
"""

# Chages the divide behavior to not do integer division
from __future__ import division

# Import Python libraries
import logging

# Import Pyomo Libraries
from pyomo.environ import Constraint, Expression, Param, \
                          PositiveReals, Reals, Set, value, Var
from pyomo.opt import SolverFactory, TerminationCondition

# Import IDAES cores
from idaes.core import declare_process_block_class, ProcessBlock, \
                        PropertyBlockDataBase, PropertyParameterBase
from idaes.core.util.misc import add_object_ref, solve_indexed_blocks

# Some more inforation about this module
__author__ = "Andrew Lee"


# Set up logger
logger = logging.getLogger('idaes.unit_model.properties')


@declare_process_block_class("PropertyParameterBlock")
class _PropertyParameterBlock(PropertyParameterBase):
    """
    Property Parameter Block Class

    Contains parameters and indexing sets associated with properties for
    water.

    """

    def build(self):
        '''
        Callable method for Block construction.
        '''
        super(_PropertyParameterBlock, self).build()

        self._make_params()

    def _make_params(self):
        ''' This section is for parameters needed for the property models.'''
        # List of valid phases in property package
        self.phase_list = Set(initialize=['Liq'])

        # Component list - a list of component identifiers
        self.component_list = Set(initialize=['H2O'])

        # Molar heat capacity
        self.cp_mol = Param(initialize=75.4,
                            within=Reals,
                            mutable=True,
                            doc="Molar heat capacity [J/mol.K]")

        # Molar density
        self.dens_mol = Param(initialize=998.0,
                              within=Reals,
                              mutable=True,
                              doc="Molar density [mol/m^3]")

        # Thermodynamic reference state
        self.temperature_ref = Param(within=PositiveReals,
                                     mutable=True,
                                     default=303.15,
                                     doc='Reference temperature [K]')

    def get_supported_properties(self):
        return {'flow_mol': {'method': None, 'units': 'mol/s'},
                'flow_mol_comp': {'method': None, 'units': 'mol/s'},
                'enth_mol': {'method': None, 'units': 'J/mol'},
                'pressure': {'method': None, 'units': 'Pa'},
                'temperature': {'method': None, 'units': 'K'},
                'cp_mol': {'method': None, 'units': 'J/mol.K'},
                'dens_mol': {'method': None, 'units': 'mol/m^3'}}

    def get_package_units(self):
        return {'time': 's',
                'length': 'm',
                'mass': 'g',
                'amount': 'mol',
                'temperature': 'K',
                'energy': 'J',
                'holdup': 'mol'}


class _PropertyBlock(ProcessBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """
    def initialize(blk, flow_mol=None, pressure=None, temperature=None,
                   hold_state=False, outlvl=0,
                   solver='ipopt', optarg={'tol': 1e-8}):
        '''
        Initialisation routine for property package.

        Keyword Arguments:
            flow_mol_comp : value at which to initialize component flows
                            (default=None)
            pressure : value at which to initialize pressure (default=None)
            temperature : value at which to initialize temperature
                          (default=None)
            outlvl : sets output level of initialisation routine

                     * 0 = no output (default)
                     * 1 = return solver state for each step in routine
                     * 2 = include solver output infomation (tee=True)

            optarg : solver options dictionary object (default=None)
            solver : str indicating whcih solver to use during
                     initialization (default = 'ipopt')
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         - True - states varaibles are not unfixed, and
                                 a dict of returned containing flags for
                                 which states were fixed during
                                 initialization.
                        - False - state variables are unfixed after
                                 initialization by calling the
                                 relase_state method

        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        '''
        # Fix state variables if not already fixed
        Fflag = {}
        Pflag = {}
        Tflag = {}

        for k in blk.keys():
            if blk[k].flow_mol.fixed is True:
                Fflag[k] = True
            else:
                Fflag[k] = False
                if flow_mol is None:
                    blk[k].flow_mol.fix(100)
                else:
                    blk[k].flow_mol.fix(flow_mol)

            if blk[k].pressure.fixed is True:
                Pflag[k] = True
            else:
                Pflag[k] = False
                if pressure is None:
                    blk[k].pressure.fix(101325.0)
                else:
                    blk[k].pressure.fix(pressure)

            if blk[k].temperature.fixed is True:
                Tflag[k] = True
            else:
                Tflag[k] = False
                if temperature is None:
                    blk[k].temperature.fix(303.15)
                else:
                    blk[k].temperature.fix(temperature)

        # Set solver options
        if outlvl > 1:
            stee = True
        else:
            stee = False

        opt = SolverFactory(solver)
        opt.options = optarg

        # ---------------------------------------------------------------------
        # Initialise values
        for k in blk.keys():
            blk[k].enth_mol == value(blk[k].cp_mol *
                                     (blk[k].temperature -
                                      blk[k].temperature_ref))

        results = solve_indexed_blocks(opt, blk, tee=stee)

        if outlvl > 0:
            if results.solver.termination_condition \
                    == TerminationCondition.optimal:
                logger.info('{} Initialisation Step 1 Complete.'
                            .format(blk.name))
            else:
                logger.warning('{} Initialisation Step 1 Failed.'
                               .format(blk.name))

        # ---------------------------------------------------------------------
        # If input block, return flags, else release state
        flags = {"Fflag": Fflag, "Pflag": Pflag, "Tflag": Tflag}

        if outlvl > 0:
            if outlvl > 0:
                logger.info('{} Initialisation Complete.'.format(blk.name))

        if hold_state is True:
            return flags
        else:
            blk.release_state(flags)

    def release_state(blk, flags, outlvl=0):
        '''
        Method to relase state variables fixed during initialisation.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of of logging
        '''
        # Unfix state variables
        for k in blk.keys():
            if flags['Fflag'][k] is False:
                blk[k].flow_mol.unfix()
            if flags['Pflag'][k] is False:
                blk[k].pressure.unfix()
            if flags['Tflag'][k] is False:
                blk[k].temperature.unfix()

        if outlvl > 0:
            if outlvl > 0:
                logger.info('{} State Released.'.format(blk.name))


@declare_process_block_class("PropertyBlock",
                             block_class=_PropertyBlock)
class PropertyBlockData(PropertyBlockDataBase):
    """
    Example property package for water
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super(PropertyBlockData, self).build()
        self._make_params()
        self._make_vars()
        self._make_constraints()
        self._make_balance_terms()

    def _make_params(self):
        '''
        This section makes references to the necessary parameters contained
        within the Property Parameter Block provided.
        '''
        # List of valid phases in property package
        add_object_ref(self, "phase_list", self.config.parameters.phase_list)

        # Component list - a list of component identifiers
        add_object_ref(self, "component_list",
                       self.config.parameters.component_list)

        # Mixture heat capacity
        add_object_ref(self, "cp_mol",
                       self.config.parameters.cp_mol)

        # Molar density
        add_object_ref(self, "dens_mol",
                       self.config.parameters.dens_mol)

        # Thermodynamic reference state
        add_object_ref(self, "temperature_ref",
                       self.config.parameters.temperature_ref)

    def _make_vars(self):
        # Create state variables
        self.flow_mol = Var(domain=Reals,
                            initialize=0.0,
                            bounds=(1e-2, 1e5),
                            doc='Total molar flowrate [mol/s]')
        self.pressure = Var(domain=Reals,
                            initialize=101325.0,
                            bounds=(1e5, 5e5),
                            doc='State pressure [Pa]')
        self.temperature = Var(domain=Reals,
                               initialize=303.15,
                               bounds=(298.15, 500.0),
                               doc='State temperature [K]')

        self.enth_mol = Var(domain=Reals,
                            initialize=0.0,
                            doc='Mixture specific enthalpy [J/mol]')

        def flow_mol_comp(b, j):
            return b.flow_mol
        self.flow_mol_comp = Expression(self.component_list,
                                        rule=flow_mol_comp)

    def _make_constraints(self):
        # Mixture molar enthalpy
        self.enth_mol_correlation = Constraint(
                expr=self.enth_mol ==
                self.cp_mol*(self.temperature-self.temperature_ref))

    def _make_balance_terms(self):
        def material_balance_term(b, i, j):
            return b.flow_mol
        self.material_balance_term = Expression(
                                        self.phase_list,
                                        self.component_list,
                                        rule=material_balance_term)

        def energy_balance_term(b, i):
            return b.enth_mol*b.flow_mol
        self.energy_balance_term = Expression(self.phase_list,
                                              rule=energy_balance_term)

        def material_density_term(b, p, j):
            return b.dens_mol
        self.material_density_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_density_term)

        def energy_density_term(b, p):
            return b.dens_mol*b.enth_mol
        self.energy_density_term = Expression(self.phase_list,
                                              rule=energy_density_term)

    def declare_port_members(b):
        members = {"flow_mol": b.flow_mol,
                   "temperature": b.temperature,
                   "pressure": b.pressure}
        return members

    def model_check(blk):
        """
        Model checks for property block
        """
        # Check temperature bounds
        if value(blk.temperature) < blk.temperature.lb:
            logger.error('{} Temperature set below lower bound.'
                         .format(blk.name))
        if value(blk.temperature) > blk.temperature.ub:
            logger.error('{} Temperature set above upper bound.'
                         .format(blk.name))

        # Check pressure bounds
        if value(blk.pressure) < blk.pressure.lb:
            logger.error('{} Pressure set below lower bound.'.format(blk.name))
        if value(blk.pressure) > blk.pressure.ub:
            logger.error('{} Pressure set above upper bound.'.format(blk.name))
