##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
MEA absorber model version 1.  This is a starting 
point to build more general detailed solvent models.
"""
from __future__ import division
from __future__ import print_function
__author__ = "John Eslick"
__version__ = "1.0.0"

# stdlib
import math
import time

# third-party
from pyomo.environ import *
from pyomo.opt import SolverFactory
from pyomo.dae.plugins.finitedifference import \
    Finite_Difference_Transformation
from pyomo.opt import SolverStatus, TerminationCondition
import pandas as pd
from IPython.display import HTML

# local
import idaes_models.core.model_util as model_util
import idaes_models.core.idaes_framework
from idaes.core.flowsheet_model import FlowsheetModel
from idaes_models.core.report import FlowsheetOutput
from absorber import Absorber
from inlet import OnePhaseInlet

R_gas = 8.314472

class AbsorberSheet(FlowsheetModel):
    def __init__(self, name, solver='ipopt', nfe=60):
        """
        Initialize the single MEA absorber flowsheet
        """
        FlowsheetModel.__init__(self, name, solver)
        self.nfe = nfe
        self._reg_initialize(
            'default', func=self._default_init, 
            doc='Default flowsheet initialization')
        self._reg_initialize(
            'json', func=self._load_json_init, 
            doc='Initialize from saved json')
        self.build()
        
    def _default_init(self, 
                      xin={'CO2':0.757, 'H2O':89.22, 'MEA':10.03},
                      yin={'CO2':10.89, 'H2O':9.245, 
                           'MEA':0.000, 'N2':79.86},
                      P=108120, Tv_in=319, Tl_in=318, 
                      uv=0.9809, ul=0.002748, 
                      L=6.069, D=0.64):
        model = self.model
        # Normalize x and y to make sure sum = 1
        xin_tot = sum(xin[i] for i in xin)
        yin_tot = sum(yin[i] for i in yin)
        for i in xin:
            xin[i] = xin[i]/xin_tot
        for i in yin:
            yin[i] = yin[i]/yin_tot
        # Set initial guess values for absorber
        model.absorb.L.value = L
        model.absorb.D.value = D
        for z in model.absorb.z:
            model.absorb.ul[z].value = ul
            model.absorb.uv[z].value = uv
            model.absorb.P[z].value = P
            model.absorb.T_l[z].value = Tl_in
            model.absorb.T_v[z].value = Tv_in
            model_util.dict_set(
                model.absorb.y_l, xin, (model.absorb.z,),(z,))
            model_util.dict_set(
                model.absorb.y_v, yin, (model.absorb.z,),(z,))
        self.units["absorb"].initialize(
            'default', self.model, solver='ipopt', h_steps=[
            0.1, 
            0.44,
            0.443,
            0.4435,
            0.4440,
            0.4445,
            0.4450,
            0.4455,
            0.4460,
            0.4465,
            0.4470,
            0.4475,
            0.448,
            0.45,
            1.0])
        
    def _load_json_init(self, fname):
        self.load_json(fname=fname)
    
    def _load_external_properties(self, p_lib="phys_prop.so"):
        '''
        Load the external property functions used in the models.
        '''
        model = self.model
        model.p_cp_ig = ExternalFunction(
            library=p_lib, function="cp_ig")
        model.p_cp_liq = ExternalFunction(
            library=p_lib, function="cp_liq")
        model.p_mw = ExternalFunction(
            library=p_lib, function="mw_ave")
        model.p_V_vap = ExternalFunction(
            library=p_lib, function="V_vap")
        model.p_V_liq = ExternalFunction(
            library=p_lib, function="V_liq")
        model.p_rho_vap = ExternalFunction(
            library=p_lib, function="rho_vap")
        model.p_rho_liq = ExternalFunction(
            library=p_lib, function="rho_liq")
        model.p_mu_vap = ExternalFunction(
            library=p_lib, function="mu_vap")
        model.p_mu_liq = ExternalFunction(
            library=p_lib, function="mu_liq")
        model.p_D_vap = ExternalFunction(
            library=p_lib, function="D_vap")
        model.p_sig = ExternalFunction(
            library=p_lib, function="surface_tension")
        model.p_vapp = ExternalFunction(
            library=p_lib, function="vapor_pressure")
        model.p_henry_co2 = ExternalFunction(
            library=p_lib, function="henry_co2")
        model.p_hvap_H2O = ExternalFunction(
            library=p_lib, function="hvap_H2O")
        model.p_hvap_MEA = ExternalFunction(
            library=p_lib, function="hvap_MEA")
        model.p_habs_CO2 = ExternalFunction(
            library=p_lib, function="habs_CO2")
        model.p_D_CO2 = ExternalFunction(
            library=p_lib, function="D_CO2")
        model.p_D_MEA = ExternalFunction(
            library=p_lib, function="D_MEA")
        model.p_D_MEACOO = ExternalFunction(
            library=p_lib, function="D_MEACOO")    
        model.p_logKeq = ExternalFunction(
            library=p_lib, function="logKeq")
        model.p_Keq = ExternalFunction(
            library=p_lib, function="Keq")
        
    def build(self):
        """
        Put the flowsheet together. Spearate from __init__ in case you
        want to do more setup before building the model.  Mabe combine
        with __init__ is better?
        """
        model = self.model
        self._load_external_properties()
        # Make absorber and get component lists
        absorb = self.add_unit(
            name="absorb", o=Absorber(nfe=self.nfe, main_model=model))
        lcomp = absorb.pyomo_block.lcomp
        comp = absorb.pyomo_block.lcomp
        # Make inlets
        liq_in = self.add_unit(
            name="liq_inlet", o=OnePhaseInlet(lcomp))
        #make connection
        self.connect("s1", liq_in.out, absorb.liq_in)
        self.expand_streams()
        
class MeaFlowsheetOutput(FlowsheetOutput):
    SECTIONS = {'location': 'Z Location',
                'temperature': 'Temperature profiles',
                'co2': 'CO2 profiles'}

    def header(self):
        m = self._model
        return HTML('''<h1>Flowsheet</h1>
        <h2>Flows</h2>
        <ul>
        <li>Area: {area} m2</li>
        <li>Liquid in: {liq_in} mol/s</li>
        <li>Vapor in: {vap_in} mol/s</li>
        <li>CO2 Capture: {co2}</li>
        </ul>
        '''.format(area=m.absorb.A.value,
                   liq_in=m.absorb.F_liq_in.value,
                   vap_in=m.absorb.F_vap_in.value,
                   co2=(1 - m.absorb.Cu_v[0, 'CO2'].value / m.absorb.Cu_v[1, 'CO2'].value)))

    def footer(self):
        return ''

    def location_table(self, style=True):
        data = [(i, z, self._model.absorb.L.value * z)
                for i, z in enumerate(self._model.absorb.z)]
        df = pd.DataFrame(data=data,  columns=('n', 'z', 'z*L'))
        if style:
            # note: 'n' is ok as-is
            return self.style_table(df, {'z': '{:8.4f}', 'z*L': '{:8.4f}'})
        else:
            return df

    def location_plot(self, **kw):
        df = self.location_table(style=False)
        return df.plot(x='z', y='z*L', kind='scatter', **kw)

    def temperature_table(self, style=True):
        data = [(self._model.absorb.T_l[z].value,
                 self._model.absorb.T_v[z].value,
                 self._model.absorb.deltaT[z].value)
                for z in self._model.absorb.z]
        df = pd.DataFrame(data=data, columns=('T_l', 'T_v', 'dT'))
        if style:
            return self.style_table(df, {
                'T_l': '{:8.4f}', 'T_v': '{:8.4f}', 'dT': '{:8.4f}'})
        else:
            return df

    def temperature_plot(self, **kw):
        df = self.temperature_table(style=False)
        return df.plot(x='T_l', y='T_v', kind='scatter', **kw)

    def co2_table(self, style=True):
        m = self._model
        data =[(
            m.absorb.prop_l_bulk.y[z,'CO2'].value,
            m.absorb.prop_v_bulk.y[z,'CO2'].value,
            m.absorb.y_star[z, 'CO2'].value,
            m.absorb.dy[z, 'CO2'].value,
            m.absorb.prop_l_bulk.alpha[z].value,
            math.exp(m.absorb.logC[z,'CO2'].value),
            m.absorb.prop_l_bulk.henry[z,'CO2'].value,
            m.absorb.prop_l_bulk.henry[z,'CO2'].value / R_gas
                     /m.absorb.prop_l_bulk.T[z].value)
               for z in m.absorb.z]
        df = pd.DataFrame(data=data,
                          columns=['x_CO2', 'y_CO2', 'y*', 'y-y*', 'alpha',
                                   'C*_LCO2', 'He', 'He/R/T'])
        if style:
            return self.style_table(df, {
                'x_CO2':'{:7.5f}', 'y_CO2': '{:7.5f}', 'y*': '{:10.3e}',
                'y-y*': '{:7.5f}', 'alpha': '{:7.5f}', 'C*_LCO2': '{:12.5e}',
                'He': '{:7.1f}', 'He/R/T': '{:7.4f}'})
        else:
            return df

