##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
MEA absorber model version 1.  This is a starting 
point to build more general detailed solvent models.
"""
from __future__ import division
from __future__ import print_function
__author__ = "John Eslick"
__version__ = "1.0.0"


import logging

# third-party
# TODO: fix to import only symbols we need
from pyomo.environ import *
from pyomo.core.base.symbol_map import *
try:
    from idaes_models.core.viewer import *
except:
    pass

# local
from abs_fs_model import AbsorberSheet, MeaFlowsheetOutput
import idaes_models.core.model_util as model_util

_log = logging.getLogger(__name__)

def init(filename=None, solver='ipopt'):
    """Create new flowsheet from provided JSON file, or by running the
    default initialization procedures.

    Returns:
        New flowsheet object
    """
    flowsheet = AbsorberSheet(name='Absorber', solver=solver, nfe=60)
    flowsheet.strip_bounds()
    if filename:
        _log.info("Loading model state from file '{}'".format(filename))
        flowsheet.initialize('json', filename)
    else:
        _log.info("Initialize model")
        flowsheet.initialize('default')
    return flowsheet

def save(flowsheet, filename):
    """Save the flowsheet

    Args:
        flowsheet: The Flowsheet object
        filename: File in which to write the JSON

    Raises:
        IOError if the file is not writable
    """
    flowsheet.save_json(fname=filename)

def run(flowsheet, check_bounds=False, keep_files=False, **extra_kw):
    """Run the flowsheet

    Args:
        flowsheet:
        check_bounds:
        keep_files:

    Returns:
        The `model` attribute of the flowsheet.
    """
    model = flowsheet.model
    
    #Do whatever to get model in proper state to solve.
    print("Model initialized now setting up problem...")
    print("Unfix velocities and set flows")
    model.absorb.ul[0].unfix()
    model.absorb.uv[1].unfix()
    print("F_liq_in = {}".format(model.absorb.F_liq_in.value))
    print("F_vap_in = {}".format(model.absorb.F_vap_in.value))
    model.absorb.F_liq_in.fix(39.12) # test case 39.12
    model.absorb.F_vap_in.fix(12.86) # test case 12.86
    print("Fixing flow rates to...")
    print("F_liq_in = {}".format(model.absorb.F_liq_in.value))
    print("F_vap_in = {}".format(model.absorb.F_vap_in.value))

    print("Solving Model... First test case")
    res_file = "test2.csv"
    if keep_files:
        model_util.get_pyomo_tmp_files()
    #flowsheet.solver_setup('ipopt')
    #flowsheet.pyomo_solver.options['-snes_monitor'] = ''
    #flowsheet.pyomo_solver.options['-ksp_monitor'] = ''
    #flowsheet.pyomo_solver.options['-snes_max_it'] = '20000'
    #flowsheet.pyomo_solver.options['-snes_max_funcs'] = '60000'
    #flowsheet.pyomo_solver.options['-snes_atol'] = '1e-10'
    #flowsheet.pyomo_solver.options['-snes_rtol'] = '1e-8'
    #flowsheet.pyomo_solver.options['-scale_eqs'] = '0'
    #flowsheet.pyomo_solver.options['-scale_vars'] = '0'
    #flowsheet.solve(tee=solverMessages, keepfiles=keepfiles)
    flowsheet.solve(tee=True)
    flowsheet.display_results() # print solver results
    flowsheet.units['absorb'].make_profile()
    print("Saving test case a profile to {}".format(res_file))
    flowsheet.units['absorb'].profile_1.to_csv(res_file)
    flowsheet.save_json('result1.json')
    
    model.absorb.F_liq_in.fix(38.00) # test case 39.12
    model.absorb.F_vap_in.fix(12.86) # test case 12.86
    print("Fixing flow rates to...")
    print("F_liq_in = {}".format(model.absorb.F_liq_in.value))
    print("F_vap_in = {}".format(model.absorb.F_vap_in.value))
    print("Solving Model... Second test case")
    res_file = "test2b.csv"
    flowsheet.solve(tee=True)
    flowsheet.display_results() # print solver results
    flowsheet.units['absorb'].make_profile()
    print("Saving test case b profile to {}".format(res_file))
    flowsheet.units['absorb'].profile_1.to_csv(res_file)
    flowsheet.save_json('result2.json')
    
    flowsheet.restore_bounds()
    flowsheet.check_bounds()
    #flowsheet.print_vars()
    
    try:
        v = QtModelViewer(flowsheet.model)
        v.show_vars()
    except:
        pass
    
    # Model output needs fixed up
    #Mass balance check
    print("\n\n**MASS BALANCE CHECK**\n")
    for i in model.absorb.lcomp:
        i_in = model.absorb.Cu_l[0, i].value \
            + model.absorb.Cu_v[1, i].value
        i_out = model.absorb.Cu_l[1, i].value \
            + model.absorb.Cu_v[0, i].value
        print("{1}_in = {0}".format(i_in, i))
        print("{1}_out = {0}".format(i_out, i))
        print("{1}_diff = {0}".format(i_in - i_out, i))
        print("{1}_pct = {0}".format((i_in - i_out)/i_in*100, i))
        print("")
    
    #model.output = MeaFlowsheetOutput(model=model,
    #                                  properties=FluidProperties)
    
    return model

if __name__ == "__main__":
    parser = model_util.create_argparse()
    args = parser.parse_args()
    flowsheet = init(filename=args.load_json)
    run(flowsheet, **vars(args))
    if args.save_json:
        save(flowsheet, args.save_json)
