##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Transport Property calculations.
"""
from __future__ import division
from __future__ import print_function

# Some more inforation about this module
__author__ = "John Eslick"
__version__ = "1.0.0"

from pyomo.environ import *
from idaes_models.core.process_base import ProcessBase
import idaes_models.core.model_util

class Transport(ProcessBase):
    """
    Class for calculating transport properties
    """
    packing_param = {
        'mellapak_250Y':{
            'a':250,      # Area (m2/m3)
            'S':0.017,    # Channel Angle (m)
            'eps':0.97,   # Porosity (m3/m3)
            'Cs':3.157,
            'CFl':2.464,
            'LpA':237,     # Not sure
            'Cv':0.357,    # Constant 
            'Cl':0.5       # Constant
        }
    } #END packing parameters
    def __init__(self, comp, lcomp, index_set, P,  T_L, y_L, T_V, y_V,
                 u_liq, u_vap, logC, ystar, packing='mellapak_250Y',
                 col_type="abs"):
        """
        Args:
        comp: component set all components in system
        lcomp: component set for components in liquid
        index_set: index_set to calculate props of discretized domain
        liq_prop: liquid property block
        vap_prop: vapor property block
        u_liq: liquid superficial velocity from main model
        u_vap: vapor superficial velocity from main model
        logC: log concentrations from chemical equilibrium
        packing: is the type of column packing
        """
        if index_set is None:
            raise Exception("Trasport block must have index_set for now")
        ProcessBase.__init__(self, index_set, index_vars=True)
        self._reg_initialize(
            'default', func=self._default_init, doc='Not sure yet')
        self.packing = packing
        self.comp = comp
        self.lcomp = lcomp
        self.P = P
        self.T_L = T_L
        self.y_L = y_L
        self.T_V = T_V
        self.y_V = y_V
        self.u_liq = u_liq
        self.u_vap = u_vap
        self.logC = logC
        self.ystar = ystar
        self.col_type = col_type
    
    def __call__(self, blk, indx=0):
        self.pyomo_block = blk
        self._create_params(blk)
        self._create_vars(blk)
        self._create_constraints(blk)
        
    def _default_init(self):
        """
        Initialization proceedure.  For this initialization, the to 
        property blocks should already be initialized.
        """
        pass #gonna rework this
        
    def _create_params(self, blk):
        """
        Create transport model parameters
        """
        pk_par = self.packing_param[self.packing]
        blk.a = Param(
            initialize=pk_par['a'], doc = "Packing area m2/m3")
        blk.eps = Param(
            initialize=pk_par['eps'], doc = "Packing void space m3/m3")
        blk.LpA = Param(
            initialize=pk_par['LpA'],
            doc = "Parameter related to packing geometry m/m2")
        blk.S = Param(
            initialize=pk_par['S'],
            doc = "Parameter related to packing geometry m")
        blk.Cv = Param(
            initialize=pk_par['Cv'], 
            doc = "Dimensionless empirical packing paramter")
        blk.Cl = Param(
            initialize=pk_par['Cl'],
            doc = "Dimensionless empirical packing paramter")
        blk.dh = Param(
            initialize=4*pk_par['eps']/pk_par['a'], 
            doc = "Hydraulic diameter (m)")
        blk.Rgas = Param(
            initialize=8.314472, doc="ideal gas constant")
        blk.g_acc = Param(
            initialize=9.8, doc="gravity acc (m/s2)")
        blk.kv_par = Param(
            initialize=(blk.Cv**2*blk.a**3)/(blk.a**2*blk.dh*blk.a**(3/2)),
            doc="parameter grouping for kv")
        blk.kl_par = Param(
            initialize=(blk.Cl**2*12**(1/3))/blk.dh,
            doc="parameter grouping for kl")
        blk.epsl_par = Param(
            initialize=(blk.LpA*blk.S**2*blk.g_acc**(2/3))**0.6471/11.45,
            doc="parameter grouping for epsl")
        blk.ae_par = Param(
            initialize=1.34*blk.a*blk.g_acc**(0.116/3)/blk.LpA**0.1547,
            doc="parameter grouping for ae")

    def _create_vars(self, blk):
        """
        Create transport model variables
        """
        blk.ae = Var(
            self.index_set, initialize=blk.a, bounds=(100, 350), 
            domain=Reals, 
            doc="Effective vapor/liquid interfaceial area (m2/m3)")
        blk.eps_l = Var(
            self.index_set, initialize=0.001, bounds=(0, 0.4), 
            domain=Reals, 
            doc="Volume fraction liquid per column volume, not including"\
                " packing(m2/m3)")
        blk.eps_v = Var(
            self.index_set, initialize=0.9, bounds=(0.6, 1.0), 
            domain=Reals, 
            doc="Volume fraction vapor per column volume, not including"\
                " packing(m2/m3)")
        blk.k_l_CO2 = Var(
            self.index_set, initialize=2e-2, bounds=(1e-6, 0.5), 
            domain=Reals, 
            doc="Liquid CO2 mass transfer coefficient")
        blk.k_v = Var(
            self.index_set, self.lcomp, initialize=2e-4, 
            bounds=(1e-4, 0.5), domain=Reals, 
            doc="Vapor mass transfer coefficient")
        blk.K_v = Var(
            self.index_set, self.lcomp, initialize=2e-3, 
            bounds = (1e-5, 0.5), domain=Reals,
            doc="Overall vapor phase mass transfer coefficient")
        blk.K_vCO2_1 = Var(
            self.index_set, initialize=1e-3, 
            bounds = (1e-5, 1e5), domain=Reals,
            doc="Overall vapor phase mass transfer coefficient")
        blk.K_vCO2_2 = Var(
            self.index_set, initialize=1e-3, 
            bounds = (1e-5, 1e5), domain=Reals,
            doc="Overall vapor phase mass transfer coefficient")
        blk.E = Var(
            self.index_set, initialize=100, bounds=(10, 1e6), 
            domain=Reals, doc="Enhancment factor")
        blk.kap = Var(
            self.index_set, initialize=240000, bounds=(1e3, 8e5),
            domain=Reals,
            doc="Reaction rate constand for CO2 and solvent")
        blk.koh = Var(
            self.index_set, initialize=200,
            bounds=(0.001, 1000), domain=Reals,
            doc="Reaction rate constand for CO2 and solvent")
        blk.h_vl = Var(
            self.index_set, initialize=50, domain=Reals,
            bounds=(1, 200),
            doc="Heat transfer coefficient")
        blk.kMEA = Var(
            self.index_set, initialize=10, domain=Reals,
            bounds=(-1e6, 1e6),
            doc="Rate constant MEA")
        blk.kH2O = Var(
            self.index_set, initialize=10, domain=Reals,
            bounds=(-1e6, 1e6),
            doc="Rate constant H2O")
        blk.Keq = Var(
            self.index_set, initialize=10, domain=Reals,
            bounds=(-1e6, 1e6),
            doc="K in regenerator enhancment factor calc")
            
    def _create_constraints(self, blk):        
        # Property functions
        p_rho_vap = self.main_model.p_rho_vap
        p_rho_liq = self.main_model.p_rho_liq
        p_mu_vap = self.main_model.p_mu_vap
        p_mu_liq = self.main_model.p_mu_liq
        p_D_vap = self.main_model.p_D_vap
        p_sig = self.main_model.p_sig
        p_henry_co2 = self.main_model.p_henry_co2
        p_D_CO2 = self.main_model.p_D_CO2
        p_D_MEA = self.main_model.p_D_MEA
        p_D_MEACOO = self.main_model.p_D_MEACOO
        p_V_liq = self.main_model.p_V_liq
        p_Keq = self.main_model.p_Keq
        
        # Parameters
        a = blk.a 
        eps = blk.eps
        LpA = blk.LpA
        S = blk.S 
        Cv = blk.Cv
        Cl = blk.Cl
        dh = blk.dh
        Rgas = blk.Rgas
        g = blk.g_acc
        # Variables
        ae = blk.ae
        eps_l = blk.eps_l
        eps_v = blk.eps_v
        k_l_CO2 = blk.k_l_CO2
        k_v = blk.k_v 
        K_v = blk.K_v
        E = blk.E
        kap = blk.kap
        koh = blk.koh
        kMEA = blk.kMEA
        kH2O = blk.kH2O
        h_vl = blk.h_vl
        y = self.y_V
        yl = self.y_L
        Keq = blk.Keq

        # Constraints part 1
        blk.eq_ae = Constraint(self.index_set, noruleinit=True)
        blk.eq_eps_l = Constraint(self.index_set, noruleinit=True)
        blk.eq_eps_v = Constraint(self.index_set, noruleinit=True)
        blk.eq_k_l_CO2 = Constraint(self.index_set, noruleinit=True)
        blk.eq_k_v = Constraint(
            self.index_set, self.lcomp, noruleinit=True)
        blk.eq_K_v = Constraint(
            self.index_set, self.lcomp, noruleinit=True)
        blk.eq_K_vCO2_1 = Constraint(self.index_set, noruleinit=True)
        blk.eq_K_vCO2_2 = Constraint(self.index_set, noruleinit=True)
        blk.eq_E = Constraint(self.index_set, noruleinit=True)
        blk.eq_kap = Constraint(self.index_set, noruleinit=True)
        blk.eq_koh = Constraint(self.index_set, noruleinit=True)
        blk.eq_kH2O = Constraint(self.index_set, noruleinit=True)
        blk.eq_kMEA = Constraint(self.index_set, noruleinit=True)
        blk.eq_h_vl = Constraint(self.index_set, noruleinit=True)
        blk.eq_Keq = Constraint(self.index_set, noruleinit=True)
        # Constraints part 2
        for z in self.index_set:
            ul = self.u_liq[z]
            uv = self.u_vap[z]
            T_L = self.T_L[z] #Liquid temperature (K)
            T_V = self.T_V[z] #Vapor temperature (K)
            P = self.P[z]
            ci = {'CO2':0.0, 'H2O':1.0, 'N2':2.0, 'MEA':3.0, 'O2':4.0}
            for i in self.lcomp:
                blk.eq_k_v.add(
                    (z,i), expr=k_v[z,i]**2*eps_v[z]*
                    p_mu_vap(T_V, y[z,'CO2'], y[z,'H2O'], y[z,'N2'], y[z,'MEA'])**(5/6) ==
                    blk.kv_par*uv**(3/2)*
                    p_D_vap(ci[i], T_V, P, y[z,'CO2'], y[z,'H2O'], y[z,'N2'], y[z,'MEA'])**(4/3)*
                    p_rho_vap(T_V, P, y[z,'CO2'], y[z,'H2O'], y[z,'N2'], y[z,'MEA'])**(5/6))
            
            blk.eq_ae.add(
                z, expr=ae[z] == blk.ae_par*ul**(0.116*4/3)*
                p_rho_liq(T_L, yl[z,'CO2'], yl[z,'H2O'], 0.0, yl[z,'MEA'])**0.116/
                p_sig(T_L, yl[z,'CO2'], yl[z,'H2O'], yl[z,'MEA'])**0.116)
            
            blk.eq_eps_l.add(
                z, expr=eps_l[z]*blk.epsl_par ==
                ((p_mu_liq(T_L, yl[z,'CO2'], yl[z,'MEA'])/
                p_rho_liq(T_L, yl[z,'CO2'], yl[z,'H2O'], 0.0, yl[z,'MEA']))**(1/3)*
                ul)**0.6471)
                
            blk.eq_eps_v.add(
                z, expr=0 == -eps + eps_l[z] + eps_v[z])
                
            blk.eq_k_l_CO2.add(
                z, expr=k_l_CO2[z]**2*eps_l[z] == blk.kv_par*ul*
                p_D_CO2(T_L, yl[z,'CO2'], yl[z,'H2O'], 0.0, yl[z,'MEA']))
                            
            blk.eq_K_v.add((z, 'H2O'), expr=K_v[z,'H2O'] == k_v[z,'H2O'])
                
            blk.eq_K_v.add((z, 'MEA'), expr=K_v[z,'MEA'] == k_v[z,'MEA'])
            
            blk.eq_K_v.add(
                (z, 'CO2'), expr=E[z]*Rgas*T_V*k_v[z,'CO2']*k_l_CO2[z] ==
                K_v[z,'CO2']*E[z]*k_l_CO2[z]*Rgas*T_V + 
                K_v[z,'CO2']*p_henry_co2(T_L, yl[z,'H2O'], yl[z,'MEA'])*k_v[z,'CO2'])
            
            # Need to add heat transfer coefficent correlation
            blk.eq_h_vl.add(z, expr=0 == -h_vl[z] + 70)
           
            #Effectivness Factor
            if self.col_type == 'abs':
                blk.eq_E.add(
                    z, expr=E[z]**2*k_l_CO2[z]**2 == (kap[z]+koh[z])*
                    p_D_CO2(T_L, yl[z,'CO2'], yl[z,'H2O'], 0.0, yl[z,'MEA']))

                blk.eq_kap.add(z, expr=kap[z] == kMEA[z] + kH2O[z])
                
                blk.eq_koh.add(
                    z, expr=0 == -koh[z] + 
                    4.315e10*exp(-6666/T_L)*exp(self.logC[z,'OH-']))
                    
                blk.eq_kH2O.add(z, expr=kH2O[z] == 
                    4.55*exp(-3287/T_L)*yl[z,'H2O']*yl[z,'MEA']/
                    p_V_liq(T_L, yl[z,'CO2'], yl[z,'H2O'], 0.0, yl[z,'MEA'])**2)
                    
                blk.eq_kMEA.add(z, expr=kMEA[z] ==
                    4610*exp(-4412/T_L)*yl[z,'MEA']**2/
                    p_V_liq(T_L, yl[z,'CO2'], yl[z,'H2O'], 0.0, yl[z,'MEA'])**2)
            elif self.col_type == 'rgn':
                blk.eq_E.add(z, expr=E[z] == 1 +
                    (p_D_MEACOO(T_L, yl[z,'CO2'], yl[z,'MEA'])/
                        p_D_CO2(T_L, yl[z,'CO2'], yl[z,'H2O'], 0.0, yl[z,'MEA']))*
                    (Keq[z]*yl[z,'CO2']/p_V_liq(T_L, yl[z,'CO2'], yl[z,'H2O'], 0.0, yl[z,'MEA']))**0.5/
                    ((1 + 
                    2*p_D_MEACOO(T_L, yl[z,'CO2'], yl[z,'MEA'])/
                        p_D_MEA(T_L, yl[z,'CO2'], yl[z,'H2O'], 0.0, yl[z,'MEA'])*
                        (Keq[z]*yl[z,'CO2']/p_V_liq(T_L, yl[z,'CO2'], yl[z,'H2O'], 0.0, yl[z,'MEA']))**0.5)*
                        (2*yl[z,'CO2']/p_V_liq(T_L, yl[z,'CO2'], yl[z,'H2O'], 0.0, yl[z,'MEA']))**0.5))
                
                blk.eq_Keq.add(z, expr=Keq[z]==p_Keq(1,T_L)/(p_Keq(8,T_L)*p_Keq(9,T_L)))
            else:
                raise(Exception("Unknown Column Type {}".format(self.col_type)))
