##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for holdup.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import Block, Expression, Set
from pyomo.common.config import ConfigValue
from idaes.core import FlowsheetBlockData, declare_process_block_class, \
                        PropertyParameterBase, PropertyBlockDataBase, \
                        UnitBlockData, CONFIG_Base
from idaes.core.holdup import HoldupData

# _get_indexing_Sets will be tested by inheriting Holdup classes.

# =============================================================================
# General test classes
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("Unit")
class _UnitData(UnitBlockData):
    CONFIG = CONFIG_Base()
    CONFIG.declare("property_package",
                   ConfigValue(default=None))
    CONFIG.declare("property_package_args",
                   ConfigValue(default={}))

    def build(self):
        super(_UnitData, self).build()


@declare_process_block_class("ParameterBlock")
class _ParameterBlock(PropertyParameterBase):
    def build(self):
        super(_ParameterBlock, self).build()
        self.phase_list = Set(initialize=["phase1", "phase2"])
        self.component_list = Set(initialize=["comp1", "comp2"])


@declare_process_block_class("PropertyBlock")
class PropertyBlockData(PropertyBlockDataBase):
    def build(self):
        super(PropertyBlockData, self).build()
        self.phase_list = Set(initialize=["phase1", "phase2"])
        self.component_list = Set(initialize=["comp1", "comp2"])
        self.element_list = Set(initialize=["element1", "element2"])
        self.element_comp = {"comp1": {"element": 1},
                             "comp2": {"element": 2}}
        self.pressure = 1

        def material_terms(b, p, j):
            return 1
        self.material_balance_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_terms)

        def energy_terms(b, p):
            return 2
        self.energy_balance_term = Expression(self.phase_list,
                                              rule=energy_terms)

        def material_density_terms(b, p, j):
            return 3
        self.material_density_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_density_terms)

        def energy_density_terms(b, p):
            return 4
        self.energy_density_term = Expression(self.phase_list,
                                              rule=energy_density_terms)


# =============================================================================
# Testing Holdup
@declare_process_block_class("Unit_NoConfig")
class _UnitDataNoConfig(UnitBlockData):
    def build(self):
        super(_UnitDataNoConfig, self).build()


@declare_process_block_class("HoldupFrame")
class _HoldupFrame(HoldupData):
    def build(self):
        pass


@pytest.fixture()
def fs():
    """
    Build a simple flowsheet model for build tests from.
    """
    fs = Flowsheet(dynamic=False, concrete=True)
    fs.p = ParameterBlock()
    fs.u = Unit()
    fs.u.h = HoldupFrame()
    return fs


def test_config_block(fs):
    # Test that ConfigBlock has correct attributes
    assert fs.u.h.config.dynamic == 'use_parent_value'
    assert fs.u.h.config.include_holdup == 'use_parent_value'
    assert fs.u.h.config.material_balance_type == 'use_parent_value'
    assert fs.u.h.config.energy_balance_type == 'use_parent_value'
    assert fs.u.h.config.momentum_balance_type == 'use_parent_value'
    assert fs.u.h.config.has_rate_reactions == 'use_parent_value'
    assert fs.u.h.config.has_equilibrium_reactions == 'use_parent_value'
    assert fs.u.h.config.has_mass_transfer == 'use_parent_value'
    assert fs.u.h.config.has_heat_transfer == 'use_parent_value'
    assert fs.u.h.config.has_work_transfer == 'use_parent_value'
    assert fs.u.h.config.has_pressure_change == 'use_parent_value'
    assert fs.u.h.config.property_package == 'use_parent_value'
    assert fs.u.h.config.property_package_args == 'use_parent_value'


def test_config_validation_valid(fs):
    # Test that ConfigBlock validation is correct for valid arguments
    fs.u.h.config.dynamic = True
    fs.u.h.config.dynamic = False

    fs.u.h.config.include_holdup = True
    fs.u.h.config.include_holdup = False

    fs.u.h.config.material_balance_type = 'none'
    fs.u.h.config.material_balance_type = 'component_total'
    fs.u.h.config.material_balance_type = 'component_phase'
    fs.u.h.config.material_balance_type = 'element_total'

    fs.u.h.config.energy_balance_type = 'none'
    fs.u.h.config.energy_balance_type = 'total'

    fs.u.h.config.momentum_balance_type = 'none'
    fs.u.h.config.momentum_balance_type = 'total'

    fs.u.h.config.has_rate_reactions = True
    fs.u.h.config.has_rate_reactions = False

    fs.u.h.config.has_equilibrium_reactions = True
    fs.u.h.config.has_equilibrium_reactions = False

    fs.u.h.config.has_mass_transfer = True
    fs.u.h.config.has_mass_transfer = False

    fs.u.h.config.has_heat_transfer = True
    fs.u.h.config.has_heat_transfer = False

    fs.u.h.config.has_work_transfer = True
    fs.u.h.config.has_work_transfer = False

    fs.u.h.config.has_pressure_change = True
    fs.u.h.config.has_pressure_change = False

    fs.u.h.config.property_package = fs.p

    # property_package_args is currently not validated


def test_config_validation_invalid(fs):
    # Test that ConfigBlock validation is correct for invalid arguments
    with pytest.raises(ValueError):
        fs.u.h.config.dynamic = "foo"
    with pytest.raises(ValueError):
        fs.u.h.config.dynamic = 5.0
    with pytest.raises(ValueError):
        fs.u.h.config.dynamic = 5
    with pytest.raises(ValueError):
        fs.u.h.config.dynamic = [0]
    with pytest.raises(ValueError):
        fs.u.h.config.dynamic = {"foo": "bar"}

    with pytest.raises(ValueError):
        fs.u.h.config.include_holdup = "foo"
    with pytest.raises(ValueError):
        fs.u.h.config.include_holdup = 5.0
    with pytest.raises(ValueError):
        fs.u.h.config.include_holdup = 5
    with pytest.raises(ValueError):
        fs.u.h.config.include_holdup = [0]
    with pytest.raises(ValueError):
        fs.u.h.config.include_holdup = {"foo": "bar"}

    with pytest.raises(ValueError):
        fs.u.h.config.material_balance_type = "foo"
    with pytest.raises(ValueError):
        fs.u.h.config.material_balance_type = 5.0
    with pytest.raises(ValueError):
        fs.u.h.config.material_balance_type = 5
    with pytest.raises(ValueError):
        fs.u.h.config.material_balance_type = [0]
    with pytest.raises(ValueError):
        fs.u.h.config.material_balance_type = {"foo": "bar"}

    with pytest.raises(ValueError):
        fs.u.h.config.energy_balance_type = "foo"
    with pytest.raises(ValueError):
        fs.u.h.config.energy_balance_type = 5.0
    with pytest.raises(ValueError):
        fs.u.h.config.energy_balance_type = 5
    with pytest.raises(ValueError):
        fs.u.h.config.energy_balance_type = [0]
    with pytest.raises(ValueError):
        fs.u.h.config.energy_balance_type = {"foo": "bar"}

    with pytest.raises(ValueError):
        fs.u.h.config.momentum_balance_type = "foo"
    with pytest.raises(ValueError):
        fs.u.h.config.momentum_balance_type = 5.0
    with pytest.raises(ValueError):
        fs.u.h.config.momentum_balance_type = 5
    with pytest.raises(ValueError):
        fs.u.h.config.momentum_balance_type = [0]
    with pytest.raises(ValueError):
        fs.u.h.config.momentum_balance_type = {"foo": "bar"}

    with pytest.raises(ValueError):
        fs.u.h.config.has_rate_reactions = "foo"
    with pytest.raises(ValueError):
        fs.u.h.config.has_rate_reactions = 5.0
    with pytest.raises(ValueError):
        fs.u.h.config.has_rate_reactions = 5
    with pytest.raises(ValueError):
        fs.u.h.config.has_rate_reactions = [0]
    with pytest.raises(ValueError):
        fs.u.h.config.has_rate_reactions = {"foo": "bar"}

    with pytest.raises(ValueError):
        fs.u.h.config.has_equilibrium_reactions = "foo"
    with pytest.raises(ValueError):
        fs.u.h.config.has_equilibrium_reactions = 5.0
    with pytest.raises(ValueError):
        fs.u.h.config.has_equilibrium_reactions = 5
    with pytest.raises(ValueError):
        fs.u.h.config.has_equilibrium_reactions = [0]
    with pytest.raises(ValueError):
        fs.u.h.config.has_equilibrium_reactions = {"foo": "bar"}

    with pytest.raises(ValueError):
        fs.u.h.config.has_mass_transfer = "foo"
    with pytest.raises(ValueError):
        fs.u.h.config.has_mass_transfer = 5.0
    with pytest.raises(ValueError):
        fs.u.h.config.has_mass_transfer = 5
    with pytest.raises(ValueError):
        fs.u.h.config.has_mass_transfer = [0]
    with pytest.raises(ValueError):
        fs.u.h.config.has_mass_transfer = {"foo": "bar"}

    with pytest.raises(ValueError):
        fs.u.h.config.has_heat_transfer = "foo"
    with pytest.raises(ValueError):
        fs.u.h.config.has_heat_transfer = 5.0
    with pytest.raises(ValueError):
        fs.u.h.config.has_heat_transfer = 5
    with pytest.raises(ValueError):
        fs.u.h.config.has_heat_transfer = [0]
    with pytest.raises(ValueError):
        fs.u.h.config.has_heat_transfer = {"foo": "bar"}

    with pytest.raises(ValueError):
        fs.u.h.config.has_work_transfer = "foo"
    with pytest.raises(ValueError):
        fs.u.h.config.has_work_transfer = 5.0
    with pytest.raises(ValueError):
        fs.u.h.config.has_work_transfer = 5
    with pytest.raises(ValueError):
        fs.u.h.config.has_work_transfer = [0]
    with pytest.raises(ValueError):
        fs.u.h.config.has_work_transfer = {"foo": "bar"}

    with pytest.raises(ValueError):
        fs.u.h.config.has_pressure_change = "foo"
    with pytest.raises(ValueError):
        fs.u.h.config.has_pressure_change = 5.0
    with pytest.raises(ValueError):
        fs.u.h.config.has_pressure_change = 5
    with pytest.raises(ValueError):
        fs.u.h.config.has_pressure_change = [0]
    with pytest.raises(ValueError):
        fs.u.h.config.has_pressure_change = {"foo": "bar"}

    with pytest.raises(ValueError):
        fs.u.h.config.property_package = "foo"
    with pytest.raises(ValueError):
        fs.u.h.config.property_package = 5.0
    with pytest.raises(ValueError):
        fs.u.h.config.property_package = 5
    with pytest.raises(ValueError):
        fs.u.h.config.property_package = [0]
    with pytest.raises(ValueError):
        fs.u.h.config.property_package = {"foo": "bar"}

    # property_package_args is currently not validated


def test_inherit_default_build_arguments(fs):
    # Test that inherit build arguments method works correctly
    fs.u.config.include_holdup = False
    fs.u.config.material_balance_type = 'element_total'
    fs.u.config.energy_balance_type = 'none'
    fs.u.config.momentum_balance_type = 'none'
    fs.u.config.has_rate_reactions = True
    fs.u.config.has_equilibrium_reactions = True
    fs.u.config.has_mass_transfer = True
    fs.u.config.has_heat_transfer = True
    fs.u.config.has_work_transfer = True
    fs.u.config.has_pressure_change = True
    fs.u.config.property_package = fs.p
    fs.u.config.property_package_args = {"foo": "bar"}

    fs.u.h._inherit_default_build_arguments()

    assert fs.u.h.config.include_holdup is False
    assert fs.u.h.config.material_balance_type == 'element_total'
    assert fs.u.h.config.energy_balance_type == 'none'
    assert fs.u.h.config.momentum_balance_type == 'none'
    assert fs.u.h.config.has_rate_reactions is True
    assert fs.u.h.config.has_equilibrium_reactions is True
    assert fs.u.h.config.has_mass_transfer is True
    assert fs.u.h.config.has_heat_transfer is True
    assert fs.u.h.config.has_work_transfer is True
    assert fs.u.h.config.has_pressure_change is True
    assert fs.u.h.config.property_package == fs.p
    assert fs.u.h.config.property_package_args == {"foo": "bar"}


def test_inherit_default_build_arguments_defaults():
    # Test that inherit build arguments method works correctly
    fs = Flowsheet(dynamic=False, concrete=True)
    fs.u = Unit_NoConfig()
    fs.u.h = HoldupFrame()
    fs.u.h._inherit_default_build_arguments()

    assert fs.u.h.config.include_holdup is True
    assert fs.u.h.config.material_balance_type == 'component_phase'
    assert fs.u.h.config.energy_balance_type == 'total'
    assert fs.u.h.config.momentum_balance_type == 'total'
    assert fs.u.h.config.has_rate_reactions is False
    assert fs.u.h.config.has_equilibrium_reactions is False
    assert fs.u.h.config.has_mass_transfer is False
    assert fs.u.h.config.has_heat_transfer is False
    assert fs.u.h.config.has_work_transfer is False
    assert fs.u.h.config.has_pressure_change is False
    assert fs.u.h.config.property_package is None
    assert fs.u.h.config.property_package_args == {}


def test_setup_dynamics_use_parent_value(fs):
    # Test that dynamic = 'use_parent_value' works correctly
    fs.u.h.config.dynamic = 'use_parent_value'

    fs.u.h._setup_dynamics()

    assert fs.u.h.config.dynamic is False
    assert fs.u.h.time == [0]


def test_setup_dynamics_use_parent_value_fail_no_dynamic():
    # Test that dynamic = 'use_parent_value' works correctly
    fs = Flowsheet(dynamic=False, concrete=True)

    # Create a Block (with no dynamic attribute)
    fs.b = Block()
    # Add a time attribute to make sure the correct failure triggers
    fs.b.time = Set(initialize=[0])

    fs.b.h = HoldupFrame()
    fs.b.h.config.dynamic = 'use_parent_value'

    # _setup_dynamics should return AttributeError
    with pytest.raises(AttributeError):
        fs.b.h._setup_dynamics()


def test_setup_dynamics_use_parent_value_fail_no_time():
    # Test that methods fails when no time domain in parent
    fs = Flowsheet(dynamic=False, concrete=True)

    # Use a Holdup for the unit model as it has a ConfigBlock
    fs.u = HoldupFrame()
    # Set the config flag
    fs.u.config.dynamic = False

    # Add a second holdup block
    fs.u.h = HoldupFrame()

    # _setup_dynamics should return AttributeError, as fs.u has no time
    with pytest.raises(AttributeError):
        fs.u.h._setup_dynamics()


def test_setup_dynamics_use_parent_value_include_holdup():
    # Test that include_holdup flag is checked properly
    fs = Flowsheet(dynamic=True, concrete=True)
    fs.u = Unit()

    fs.u.h = HoldupFrame()
    fs.u.h.config.include_holdup = False

    fs.u.h._setup_dynamics()

    # Check that include_holdup has been changed to True
    assert fs.u.h.config.include_holdup is True


def test_get_property_package_specified(fs):
    # Test that get_property_package gets module information when specified
    fs.u.h.config.property_package = fs.p

    fs.u.h.get_property_package()

    assert fs.u.h.property_module == fs.p.property_module


def test_get_property_package_default(fs):
    # Test that get_property_package finds default package information
    fs.config.default_property_package = fs.p
    fs.u.h.config.property_package = None

    fs.u.h.get_property_package()

    assert fs.u.h.property_module == fs.p.property_module


def test_get_property_package_default_fails(fs):
    # Test that get_property_package fails when default package not set
    # TODO : Need more checks for other failure modes
    fs.u.h.config.property_package = None

    with pytest.raises(ValueError):
        fs.u.h.get_property_package()


def test_get_property_package_collect_args(fs):
    # Test that get_property_package collects build args correctly
    fs.u.h.config.property_package = fs.p
    fs.u.h.config.property_package_args = {"foo": 1, "bar": 2}

    fs.p.config.default_arguments = {"foo": 3, "baz": 4}

    fs.u.h.get_property_package()

    assert fs.u.h.config.property_package_args == {"foo": 1,
                                                   "bar": 2,
                                                   "baz": 4}


def test_build(fs):
    # Test that build method calls all submethods
    fs.u.h.config.property_package = fs.p

    super(_HoldupFrame, fs.u.h).build()

    assert fs.u.h.time == fs.time
    assert fs.u.h.property_module == fs.p.property_module
    assert fs.u.h.config.material_balance_type == 'component_phase'


# =============================================================================
# Test ConfigBase
def test_config_base():
    # Test that Config_Base has correct attributes
    config = CONFIG_Base()

    assert config.dynamic == 'use_parent_value'
    assert not config.include_holdup
    assert config.material_balance_type == 'component_phase'
    assert config.energy_balance_type == 'total'
    assert config.momentum_balance_type == 'total'
    assert not config.has_rate_reactions
    assert not config.has_equilibrium_reactions
    assert not config.has_mass_transfer
    assert not config.has_heat_transfer
    assert not config.has_work_transfer
    assert not config.has_pressure_change


def test_config_base_validation_valid(fs):
    # Test that ConfigBlock validation is correct for valid arguments
    config = CONFIG_Base()

    config.dynamic = True
    config.dynamic = False

    config.include_holdup = True
    config.include_holdup = False

    config.material_balance_type = 'none'
    config.material_balance_type = 'component_total'
    config.material_balance_type = 'component_phase'
    config.material_balance_type = 'element_total'

    config.energy_balance_type = 'none'
    config.energy_balance_type = 'total'

    config.momentum_balance_type = 'none'
    config.momentum_balance_type = 'total'

    config.has_rate_reactions = True
    config.has_rate_reactions = False

    config.has_equilibrium_reactions = True
    config.has_equilibrium_reactions = False

    config.has_mass_transfer = True
    config.has_mass_transfer = False

    config.has_heat_transfer = True
    config.has_heat_transfer = False

    config.has_work_transfer = True
    config.has_work_transfer = False

    config.has_pressure_change = True
    config.has_pressure_change = False


def test_config_base_validation_invalid(fs):
    # Test that ConfigBlock validation is correct for invalid arguments
    config = CONFIG_Base()

    with pytest.raises(ValueError):
        config.dynamic = "foo"
    with pytest.raises(ValueError):
        config.dynamic = 5.0
    with pytest.raises(ValueError):
        config.dynamic = 5
    with pytest.raises(ValueError):
        config.dynamic = [0]
    with pytest.raises(ValueError):
        config.dynamic = {"foo": "bar"}

    with pytest.raises(ValueError):
        config.include_holdup = "foo"
    with pytest.raises(ValueError):
        config.include_holdup = 5.0
    with pytest.raises(ValueError):
        config.include_holdup = 5
    with pytest.raises(ValueError):
        config.include_holdup = [0]
    with pytest.raises(ValueError):
        config.include_holdup = {"foo": "bar"}

    with pytest.raises(ValueError):
        config.material_balance_type = "foo"
    with pytest.raises(ValueError):
        config.material_balance_type = 5.0
    with pytest.raises(ValueError):
        config.material_balance_type = 5
    with pytest.raises(ValueError):
        config.material_balance_type = [0]
    with pytest.raises(ValueError):
        config.material_balance_type = {"foo": "bar"}

    with pytest.raises(ValueError):
        config.energy_balance_type = "foo"
    with pytest.raises(ValueError):
        config.energy_balance_type = 5.0
    with pytest.raises(ValueError):
        config.energy_balance_type = 5
    with pytest.raises(ValueError):
        config.energy_balance_type = [0]
    with pytest.raises(ValueError):
        config.energy_balance_type = {"foo": "bar"}

    with pytest.raises(ValueError):
        config.momentum_balance_type = "foo"
    with pytest.raises(ValueError):
        config.momentum_balance_type = 5.0
    with pytest.raises(ValueError):
        config.momentum_balance_type = 5
    with pytest.raises(ValueError):
        config.momentum_balance_type = [0]
    with pytest.raises(ValueError):
        config.momentum_balance_type = {"foo": "bar"}

    with pytest.raises(ValueError):
        config.has_rate_reactions = "foo"
    with pytest.raises(ValueError):
        config.has_rate_reactions = 5.0
    with pytest.raises(ValueError):
        config.has_rate_reactions = 5
    with pytest.raises(ValueError):
        config.has_rate_reactions = [0]
    with pytest.raises(ValueError):
        config.has_rate_reactions = {"foo": "bar"}

    with pytest.raises(ValueError):
        config.has_equilibrium_reactions = "foo"
    with pytest.raises(ValueError):
        config.has_equilibrium_reactions = 5.0
    with pytest.raises(ValueError):
        config.has_equilibrium_reactions = 5
    with pytest.raises(ValueError):
        config.has_equilibrium_reactions = [0]
    with pytest.raises(ValueError):
        config.has_equilibrium_reactions = {"foo": "bar"}

    with pytest.raises(ValueError):
        config.has_mass_transfer = "foo"
    with pytest.raises(ValueError):
        config.has_mass_transfer = 5.0
    with pytest.raises(ValueError):
        config.has_mass_transfer = 5
    with pytest.raises(ValueError):
        config.has_mass_transfer = [0]
    with pytest.raises(ValueError):
        config.has_mass_transfer = {"foo": "bar"}

    with pytest.raises(ValueError):
        config.has_heat_transfer = "foo"
    with pytest.raises(ValueError):
        config.has_heat_transfer = 5.0
    with pytest.raises(ValueError):
        config.has_heat_transfer = 5
    with pytest.raises(ValueError):
        config.has_heat_transfer = [0]
    with pytest.raises(ValueError):
        config.has_heat_transfer = {"foo": "bar"}

    with pytest.raises(ValueError):
        config.has_work_transfer = "foo"
    with pytest.raises(ValueError):
        config.has_work_transfer = 5.0
    with pytest.raises(ValueError):
        config.has_work_transfer = 5
    with pytest.raises(ValueError):
        config.has_work_transfer = [0]
    with pytest.raises(ValueError):
        config.has_work_transfer = {"foo": "bar"}

    with pytest.raises(ValueError):
        config.has_pressure_change = "foo"
    with pytest.raises(ValueError):
        config.has_pressure_change = 5.0
    with pytest.raises(ValueError):
        config.has_pressure_change = 5
    with pytest.raises(ValueError):
        config.has_pressure_change = [0]
    with pytest.raises(ValueError):
        config.has_pressure_change = {"foo": "bar"}

