##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for holdup.

Author: Andrew Lee
"""
import pytest
from pyomo.environ import ConcreteModel, Expression, Set, Var
from pyomo.common.config import ConfigValue
from idaes.core import FlowsheetBlockData, declare_process_block_class, \
                        PropertyParameterBase, PropertyBlockDataBase, \
                        PropertyBlockBase, UnitBlockData
from idaes.core.holdup import Holdup0D, Holdup0dData


# -----------------------------------------------------------------------------
# Setup a basic unit model for testing
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("Unit")
class UnitData(UnitBlockData):
    def build(self):
        super(UnitData, self).build()


@declare_process_block_class("PropertyParameterBlock")
class PropertyParameterData(PropertyParameterBase):
    def build(self):
        super(PropertyParameterData, self).build()

    def get_supported_properties(self):
        return {'flow': {'method': None, 'units': 'mol/s'},
                'pressure': {'method': None, 'units': 'Pa'},
                'temperature': {'method': None, 'units': 'K'}}

    def get_package_units(self):
        return {'time': 's',
                'length': 'm',
                'mass': 'g',
                'amount': 'mol',
                'temperature': 'K',
                'energy': 'J',
                'holdup': 'mol'}


class _PropertyBlock(PropertyBlockBase):
    def initialize(blk, outlvl, solver, optarg, hold_state, a=1, b=2, c=3):
        fflag = {}
        tflag = {}
        pflag = {}

        for k in blk.keys():
            blk[k].pressure.fix(10)
            blk[k].temperature.fix(20)
            blk[k].flow.fix(30)

            pflag[k] = True
            tflag[k] = True
            fflag[k] = True

        return {"pflag": pflag, "tflag": tflag, "fflag": fflag}

    def release_state(blk, flags, outlvl):
        for k in blk.keys():
            if flags["pflag"][k]:
                blk[k].pressure.unfix()
            if flags["tflag"][k]:
                blk[k].temperature.unfix()
            if flags["fflag"][k]:
                blk[k].flow.unfix()


@declare_process_block_class("PropertyBlock",
                             block_class=_PropertyBlock)
class PropertyBlockData(PropertyBlockDataBase):
    def build(self):
        super(PropertyBlockData, self).build()
        self.phase_list = Set(initialize=["phase1", "phase2"])
        self.component_list = Set(initialize=["comp1", "comp2"])
        self.rate_reaction_idx = Set(initialize=["rate"])
        self.rate_reaction_stoichiometry = {("rate", "phase1", "comp1"): 1,
                                            ("rate", "phase1", "comp2"): 2,
                                            ("rate", "phase2", "comp1"): 3,
                                            ("rate", "phase2", "comp2"): 4}
        self.equilibrium_reaction_idx = Set(initialize=["equil"])
        self.equilibrium_reaction_stoichiometry = {
                ("equil", "phase1", "comp1"): 5,
                ("equil", "phase1", "comp2"): 6,
                ("equil", "phase2", "comp1"): 7,
                ("equil", "phase2", "comp2"): 8}
        self.phase_equilibrium_idx = Set(initialize=["pe"])
        self.phase_equilibrium_list = {"pe": ["comp1", ("phase1", "phase2")]}
        self.element_list = Set(initialize=["element"])
        self.element_comp = {"comp1": {"element": 1},
                             "comp2": {"element": 2}}
        self.pressure = Var(initialize=1.0)
        self.temperature = Var(initialize=2.0)
        self.flow = Var(self.component_list, initialize=3.0)

        def material_terms(b, p, j):
            return 1
        self.material_balance_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_terms)

        def energy_terms(b, p):
            return 2
        self.energy_balance_term = Expression(self.phase_list,
                                              rule=energy_terms)

        def material_density_terms(b, p, j):
            return 3
        self.material_density_term = Expression(self.phase_list,
                                                self.component_list,
                                                rule=material_density_terms)

        def energy_density_terms(b, p):
            return 4
        self.energy_density_term = Expression(self.phase_list,
                                              rule=energy_density_terms)

    def declare_port_members(b):
        members = {"a": b.pressure,
                   "b": b.temperature,
                   "c": b.flow}
        return members

    def model_check(b):
        b.test = True


@declare_process_block_class("Holdup0dFrame")
class _Holdup0dFrame(Holdup0dData):
    def build(self):
        super(Holdup0dData, self).build()


@pytest.fixture()
def ms():
    """
    Build a basic Unit model with one Holdup for testing.
    """
    ms = ConcreteModel()
    ms.fs = Flowsheet(dynamic=False)
    ms.fs.prop = PropertyParameterBlock()

    ms.fs.u = Unit()
    ms.fs.u.config.declare("property_package", ConfigValue(default=ms.fs.prop))
    ms.fs.u.config.declare("property_package_args", ConfigValue(default={}))

    ms.fs.u.holdup = Holdup0dFrame()

    return ms


@pytest.fixture()
def md():
    """
    Build a basic Unit model with one Holdup for testing.
    """
    md = ConcreteModel()
    md.fs = Flowsheet(dynamic=True)
    md.fs.prop = PropertyParameterBlock()

    md.fs.u = Unit()
    md.fs.u.config.declare("property_package", ConfigValue(default=md.fs.prop))
    md.fs.u.config.declare("property_package_args", ConfigValue(default={}))

    md.fs.u.holdup = Holdup0dFrame()

    return md


# =============================================================================
# Testing Holdup0D

# -----------------------------------------------------------------------------
# Testing _make_prop_blocks
def test_make_prop_blocks(ms):
    # Test that property blocks are constructed properly
    ms.fs.u.holdup._make_prop_blocks()

    assert isinstance(ms.fs.u.holdup.properties_in, PropertyBlock)
    assert isinstance(ms.fs.u.holdup.properties_out, PropertyBlock)

    assert isinstance(ms.fs.u.holdup.properties_in[0].material_balance_term,
                      Expression)
    assert isinstance(ms.fs.u.holdup.properties_out[0].material_balance_term,
                      Expression)

    assert not ms.fs.u.holdup.properties_in[0].config.has_sum_fractions
    assert ms.fs.u.holdup.properties_out[0].config.has_sum_fractions
    assert not (ms.fs.u.holdup.properties_in[0].config.
                calculate_equilibrium_reactions)
    assert not (ms.fs.u.holdup.properties_out[0].config.
                calculate_equilibrium_reactions)
    assert not (ms.fs.u.holdup.properties_in[0].config.
                calculate_phase_equilibrium)
    assert not (ms.fs.u.holdup.properties_out[0].config.
                calculate_phase_equilibrium)


def test_make_prop_blocks_has_equilibrium_reactions(ms):
    # Test that property blocks are constructed properly
    ms.fs.u.holdup.config.has_equilibrium_reactions = True
    ms.fs.u.holdup._make_prop_blocks()

    assert isinstance(ms.fs.u.holdup.properties_in, PropertyBlock)
    assert isinstance(ms.fs.u.holdup.properties_out, PropertyBlock)

    assert isinstance(ms.fs.u.holdup.properties_in[0].material_balance_term,
                      Expression)
    assert isinstance(ms.fs.u.holdup.properties_out[0].material_balance_term,
                      Expression)

    assert not ms.fs.u.holdup.properties_in[0].config.has_sum_fractions
    assert ms.fs.u.holdup.properties_out[0].config.has_sum_fractions
    assert (ms.fs.u.holdup.properties_in[0].config.
            calculate_equilibrium_reactions)
    assert (ms.fs.u.holdup.properties_out[0].config.
            calculate_equilibrium_reactions)
    assert not (ms.fs.u.holdup.properties_in[0].config.
                calculate_phase_equilibrium)
    assert not (ms.fs.u.holdup.properties_out[0].config.
                calculate_phase_equilibrium)


def test_make_prop_blocks_has_phase_equilibrium(ms):
    # Test that property blocks are constructed properly
    ms.fs.u.holdup.config.has_phase_equilibrium = True
    ms.fs.u.holdup._make_prop_blocks()

    assert isinstance(ms.fs.u.holdup.properties_in, PropertyBlock)
    assert isinstance(ms.fs.u.holdup.properties_out, PropertyBlock)

    assert isinstance(ms.fs.u.holdup.properties_in[0].material_balance_term,
                      Expression)
    assert isinstance(ms.fs.u.holdup.properties_out[0].material_balance_term,
                      Expression)

    assert not ms.fs.u.holdup.properties_in[0].config.has_sum_fractions
    assert ms.fs.u.holdup.properties_out[0].config.has_sum_fractions
    assert not (ms.fs.u.holdup.properties_in[0].config.
                calculate_equilibrium_reactions)
    assert not (ms.fs.u.holdup.properties_out[0].config.
                calculate_equilibrium_reactions)
    assert (ms.fs.u.holdup.properties_in[0].config.
            calculate_phase_equilibrium)
    assert (ms.fs.u.holdup.properties_out[0].config.
            calculate_phase_equilibrium)


# -----------------------------------------------------------------------------
# Testing _get_indexing_sets
def test_get_indexing_sets_basic(ms):
    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()

    assert (ms.fs.u.holdup.phase_list ==
            ms.fs.u.holdup.properties_out[0].phase_list)
    assert (ms.fs.u.holdup.component_list ==
            ms.fs.u.holdup.properties_out[0].component_list)
    assert not hasattr(ms.fs.u.holdup, 'rate_reaction_idx')
    assert not hasattr(ms.fs.u.holdup, 'equilibrium_reaction_idx')
    assert not hasattr(ms.fs.u.holdup, 'phase_equilibrium_idx')
    assert not hasattr(ms.fs.u.holdup, 'element_list')


def test_get_indexing_sets_rate_supported(ms):
    ms.fs.u.holdup.config.has_rate_reactions = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()

    assert (ms.fs.u.holdup.phase_list ==
            ms.fs.u.holdup.properties_out[0].phase_list)
    assert (ms.fs.u.holdup.component_list ==
            ms.fs.u.holdup.properties_out[0].component_list)
    assert not hasattr(ms.fs.u.holdup, 'equilibrium_reaction_idx')
    assert not hasattr(ms.fs.u.holdup, 'phase_equilibrium_idx')
    assert not hasattr(ms.fs.u.holdup, 'element_list')
    assert (ms.fs.u.holdup.rate_reaction_idx ==
            ms.fs.u.holdup.properties_out[0].rate_reaction_idx)
    assert (ms.fs.u.holdup.rate_reaction_stoichiometry ==
            ms.fs.u.holdup.properties_out[0].rate_reaction_stoichiometry)


def test_get_indexing_sets_rate_not_supported(ms):
    # Remove rate_reaction_idx and test correct behaviour
    ms.fs.u.holdup.config.has_rate_reactions = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup.properties_out[0].del_component(
            ms.fs.u.holdup.properties_out[0].rate_reaction_idx)

    ms.fs.u.holdup._get_indexing_sets()

    assert (ms.fs.u.holdup.phase_list ==
            ms.fs.u.holdup.properties_out[0].phase_list)
    assert (ms.fs.u.holdup.component_list ==
            ms.fs.u.holdup.properties_out[0].component_list)
    assert not hasattr(ms.fs.u.holdup, 'rate_reaction_idx')
    assert not hasattr(ms.fs.u.holdup, 'equilibrium_reaction_idx')
    assert not hasattr(ms.fs.u.holdup, 'phase_equilibrium_idx')
    assert not hasattr(ms.fs.u.holdup, 'element_list')


def test_get_indexing_sets_equilibrium_supported(ms):
    ms.fs.u.holdup.config.has_equilibrium_reactions = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()

    assert (ms.fs.u.holdup.phase_list ==
            ms.fs.u.holdup.properties_out[0].phase_list)
    assert (ms.fs.u.holdup.component_list ==
            ms.fs.u.holdup.properties_out[0].component_list)
    assert not hasattr(ms.fs.u.holdup, 'rate_reaction_idx')
    assert not hasattr(ms.fs.u.holdup, 'phase_equilibrium_idx')
    assert not hasattr(ms.fs.u.holdup, 'element_list')
    assert (ms.fs.u.holdup.equilibrium_reaction_idx ==
            ms.fs.u.holdup.properties_out[0].equilibrium_reaction_idx)
    assert (
        ms.fs.u.holdup.equilibrium_reaction_stoichiometry ==
        ms.fs.u.holdup.properties_out[0].equilibrium_reaction_stoichiometry)


def test_get_indexing_sets_equilibrium_not_supported(ms):
    # Remove rate_reaction_idx and test correct behaviour
    ms.fs.u.holdup.config.has_equilibrium_reactions = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup.properties_out[0].del_component(
            ms.fs.u.holdup.properties_out[0].equilibrium_reaction_idx)

    ms.fs.u.holdup._get_indexing_sets()

    assert (ms.fs.u.holdup.phase_list ==
            ms.fs.u.holdup.properties_out[0].phase_list)
    assert (ms.fs.u.holdup.component_list ==
            ms.fs.u.holdup.properties_out[0].component_list)
    assert not hasattr(ms.fs.u.holdup, 'rate_reaction_idx')
    assert not hasattr(ms.fs.u.holdup, 'equilibrium_reaction_idx')
    assert not hasattr(ms.fs.u.holdup, 'phase_equilibrium_idx')
    assert not hasattr(ms.fs.u.holdup, 'element_list')


def test_get_indexing_sets_phase_equilibrium_supported(ms):
    ms.fs.u.holdup.config.has_phase_equilibrium = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()

    assert (ms.fs.u.holdup.phase_list ==
            ms.fs.u.holdup.properties_out[0].phase_list)
    assert (ms.fs.u.holdup.component_list ==
            ms.fs.u.holdup.properties_out[0].component_list)
    assert not hasattr(ms.fs.u.holdup, 'rate_reaction_idx')
    assert not hasattr(ms.fs.u.holdup, 'equilibrium_reaction_idx')
    assert not hasattr(ms.fs.u.holdup, 'element_list')
    assert (ms.fs.u.holdup.phase_equilibrium_idx ==
            ms.fs.u.holdup.properties_out[0].phase_equilibrium_idx)
    assert (ms.fs.u.holdup.phase_equilibrium_list ==
            ms.fs.u.holdup.properties_out[0].phase_equilibrium_list)


def test_get_indexing_sets_phase_equilibrium_not_supported(ms):
    # Remove rate_reaction_idx and test correct behaviour
    ms.fs.u.holdup.config.has_phase_equilibrium = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup.properties_out[0].del_component(
            ms.fs.u.holdup.properties_out[0].phase_equilibrium_idx)

    ms.fs.u.holdup._get_indexing_sets()

    assert (ms.fs.u.holdup.phase_list ==
            ms.fs.u.holdup.properties_out[0].phase_list)
    assert (ms.fs.u.holdup.component_list ==
            ms.fs.u.holdup.properties_out[0].component_list)
    assert not hasattr(ms.fs.u.holdup, 'rate_reaction_idx')
    assert not hasattr(ms.fs.u.holdup, 'equilibrium_reaction_idx')
    assert not hasattr(ms.fs.u.holdup, 'phase_equilibrium_idx')
    assert not hasattr(ms.fs.u.holdup, 'element_list')


def test_get_indexing_sets_elemental(ms):
    ms.fs.u.holdup.config.material_balance_type = 'element_total'

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()

    assert hasattr(ms.fs.u.holdup, 'element_list')


def test_get_indexing_sets_elemental_not_supported(ms):
    ms.fs.u.holdup.config.material_balance_type = 'element_total'

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup.properties_out[0].del_component(
            ms.fs.u.holdup.properties_out[0].element_list)

    with pytest.raises(AttributeError):
        ms.fs.u.holdup._get_indexing_sets()


def test_get_indexing_sets_invalid_package1(ms):
    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup.properties_out[0].del_component(
            ms.fs.u.holdup.properties_out[0].phase_list)

    with pytest.raises(AttributeError):
        ms.fs.u.holdup._get_indexing_sets()


def test_get_indexing_sets_invalid_package2(ms):
    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup.properties_out[0].del_component(
            ms.fs.u.holdup.properties_out[0].component_list)

    with pytest.raises(AttributeError):
        ms.fs.u.holdup._get_indexing_sets()


# -----------------------------------------------------------------------------
# Testing _make_vars
def test_make_vars_include_holdup_true(ms):
    ms.fs.u.holdup.config.include_holdup = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()

    assert hasattr(ms.fs.u.holdup, "volume")


def test_make_vars_include_holdup_false(ms):
    ms.fs.u.holdup.config.include_holdup = False

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()

    assert not hasattr(ms.fs.u.holdup, "volume")


def test_make_vars_comp_phase_dynamic(md):
    md.fs.u.holdup.config.material_balance_type = 'component_phase'

    md.fs.u.holdup._make_prop_blocks()
    md.fs.u.holdup._get_indexing_sets()
    md.fs.u.holdup._make_vars()

    assert hasattr(md.fs.u.holdup, "material_holdup")
    assert hasattr(md.fs.u.holdup, "material_accumulation")
    assert not hasattr(md.fs.u.holdup, "element_holdup")
    assert not hasattr(md.fs.u.holdup, "element_accumulation")


def test_make_vars_comp_phase_ss_include_holdup_true(ms):
    ms.fs.u.holdup.config.include_holdup = True
    ms.fs.u.holdup.config.material_balance_type = 'component_phase'

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()

    assert hasattr(ms.fs.u.holdup, "material_holdup")
    assert not hasattr(ms.fs.u.holdup, "material_accumulation")
    assert not hasattr(ms.fs.u.holdup, "element_holdup")
    assert not hasattr(ms.fs.u.holdup, "element_accumulation")


def test_make_vars_comp_phase_ss_include_holdup_false(ms):
    ms.fs.u.holdup.config.include_holdup = False
    ms.fs.u.holdup.config.material_balance_type = 'component_phase'

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()

    assert not hasattr(ms.fs.u.holdup, "material_holdup")
    assert not hasattr(ms.fs.u.holdup, "material_accumulation")
    assert not hasattr(ms.fs.u.holdup, "element_holdup")
    assert not hasattr(ms.fs.u.holdup, "element_accumulation")


def test_make_vars_comp_total_dynamic(md):
    md.fs.u.holdup.config.material_balance_type = 'component_total'

    md.fs.u.holdup._make_prop_blocks()
    md.fs.u.holdup._get_indexing_sets()
    md.fs.u.holdup._make_vars()

    assert hasattr(md.fs.u.holdup, "material_holdup")
    assert hasattr(md.fs.u.holdup, "material_accumulation")
    assert not hasattr(md.fs.u.holdup, "element_holdup")
    assert not hasattr(md.fs.u.holdup, "element_accumulation")


def test_make_vars_comp_total_ss_inlcude_holdup_true(ms):
    ms.fs.u.holdup.config.include_holdup = True
    ms.fs.u.holdup.config.material_balance_type = 'component_total'

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()

    assert hasattr(ms.fs.u.holdup, "material_holdup")
    assert not hasattr(ms.fs.u.holdup, "material_accumulation")
    assert not hasattr(ms.fs.u.holdup, "element_holdup")
    assert not hasattr(ms.fs.u.holdup, "element_accumulation")


def test_make_vars_comp_total_ss_inlcude_holdup_false(ms):
    ms.fs.u.holdup.config.include_holdup = False
    ms.fs.u.holdup.config.material_balance_type = 'component_total'

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()

    assert not hasattr(ms.fs.u.holdup, "material_holdup")
    assert not hasattr(ms.fs.u.holdup, "material_accumulation")
    assert not hasattr(ms.fs.u.holdup, "element_holdup")
    assert not hasattr(ms.fs.u.holdup, "element_accumulation")


def test_make_vars_element_total_dynamic(md):
    md.fs.u.holdup.config.material_balance_type = 'element_total'

    md.fs.u.holdup._make_prop_blocks()
    md.fs.u.holdup._get_indexing_sets()
    md.fs.u.holdup._make_vars()

    assert not hasattr(md.fs.u.holdup, "material_holdup")
    assert not hasattr(md.fs.u.holdup, "material_accumulation")
    assert hasattr(md.fs.u.holdup, "element_holdup")
    assert hasattr(md.fs.u.holdup, "element_accumulation")


def test_make_vars_element_total_ss_inlcude_holdup_true(ms):
    ms.fs.u.holdup.config.include_holdup = True
    ms.fs.u.holdup.config.material_balance_type = 'element_total'

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()

    assert not hasattr(ms.fs.u.holdup, "material_holdup")
    assert not hasattr(ms.fs.u.holdup, "material_accumulation")
    assert hasattr(ms.fs.u.holdup, "element_holdup")
    assert not hasattr(ms.fs.u.holdup, "element_accumulation")


def test_make_vars_element_total_ss_inlcude_holdup_false(ms):
    ms.fs.u.holdup.config.include_holdup = False
    ms.fs.u.holdup.config.material_balance_type = 'element_total'

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()

    assert not hasattr(ms.fs.u.holdup, "material_holdup")
    assert not hasattr(ms.fs.u.holdup, "material_accumulation")
    assert not hasattr(ms.fs.u.holdup, "element_holdup")
    assert not hasattr(ms.fs.u.holdup, "element_accumulation")


def test_make_vars_no_material_balance_dynamic(md):
    md.fs.u.holdup.config.material_balance_type = 'none'

    md.fs.u.holdup._make_prop_blocks()
    md.fs.u.holdup._get_indexing_sets()
    md.fs.u.holdup._make_vars()

    assert not hasattr(md.fs.u.holdup, "material_holdup")
    assert not hasattr(md.fs.u.holdup, "material_accumulation")
    assert not hasattr(md.fs.u.holdup, "element_holdup")
    assert not hasattr(md.fs.u.holdup, "element_accumulation")


def test_make_vars_no_material_balance_ss(ms):
    ms.fs.u.holdup.config.material_balance_type = 'none'

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()

    assert not hasattr(ms.fs.u.holdup, "material_holdup")
    assert not hasattr(ms.fs.u.holdup, "material_accumulation")
    assert not hasattr(ms.fs.u.holdup, "element_holdup")
    assert not hasattr(ms.fs.u.holdup, "element_accumulation")


def test_make_vars_ebalance_total_dynamic(md):
    md.fs.u.holdup.config.energy_balance_type = 'total'

    md.fs.u.holdup._make_prop_blocks()
    md.fs.u.holdup._get_indexing_sets()
    md.fs.u.holdup._make_vars()

    assert hasattr(md.fs.u.holdup, "energy_holdup")
    assert hasattr(md.fs.u.holdup, "energy_accumulation")


def test_make_vars_ebalance_total_ss_include_holdup_true(ms):
    ms.fs.u.holdup.config.include_holdup = True
    ms.fs.u.holdup.config.energy_balance_type = 'total'

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()

    assert hasattr(ms.fs.u.holdup, "energy_holdup")
    assert not hasattr(ms.fs.u.holdup, "energy_accumulation")


def test_make_vars_ebalance_total_ss_include_holdup_false(ms):
    ms.fs.u.holdup.config.include_holdup = False
    ms.fs.u.holdup.config.energy_balance_type = 'total'

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()

    assert not hasattr(ms.fs.u.holdup, "energy_holdup")
    assert not hasattr(ms.fs.u.holdup, "energy_accumulation")


def test_make_vars_ebalance_none(ms):
    ms.fs.u.holdup.config.energy_balance_type = 'none'

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()

    assert not hasattr(ms.fs.u.holdup, "energy_holdup")
    assert not hasattr(ms.fs.u.holdup, "energy_accumulation")


# -----------------------------------------------------------------------------
# Testing _make_phase_frac
def test_make_phase_frac_multiphase(ms):
    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_phase_frac()

    assert hasattr(ms.fs.u.holdup, "phase_fraction")
    assert len(ms.fs.u.holdup.phase_fraction) == 2
    assert hasattr(ms.fs.u.holdup, "sum_of_phase_fractions")


def test_make_phase_frac_single_phase(ms):
    ms.fs.u.holdup.phase_list = Set(initialize=["phase1"])
    ms.fs.u.holdup._make_phase_frac()

    assert hasattr(ms.fs.u.holdup, "phase_fraction")
    assert len(ms.fs.u.holdup.phase_fraction) == 1
    assert not hasattr(ms.fs.u.holdup, "sum_of_phase_fractions")


# -----------------------------------------------------------------------------
# Testing _make_component_balances
def test_make_component_balances_total_phase_comp_list_present(ms):
    ms.fs.u.holdup.config.material_balance_type = 'component_total'
    ms.fs.prop.phase_component_list = {"phase1": ["comp1", "comp2"],
                                       "phase2": ["comp1"]}

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()

    ms.fs.u.holdup._make_component_balances()

    for k in ms.fs.u.holdup.material_balance.keys():
        assert k in [(0, "comp1"),
                     (0, "comp2")]


def test_make_component_balances_total_phase_comp_list_not_present(ms):
    ms.fs.u.holdup.config.material_balance_type = 'component_total'
    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()

    ms.fs.u.holdup._make_component_balances()

    for k in ms.fs.u.holdup.material_balance.keys():
        assert k in [(0, "comp1"),
                     (0, "comp2")]


def test_make_component_balances_phase_phase_comp_list_present(ms):
    ms.fs.u.holdup.config.material_balance_type = 'component_phase'
    ms.fs.prop.phase_component_list = {"phase1": ["comp1", "comp2"],
                                       "phase2": ["comp1"]}

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()

    ms.fs.u.holdup._make_component_balances()

    for k in ms.fs.u.holdup.material_balance.keys():
        assert k in [(0, "phase1", "comp1"),
                     (0, "phase1", "comp2"),
                     (0, "phase2", "comp1")]


def test_make_component_balances_phase_phase_comp_list_not_present(ms):
    ms.fs.u.holdup.config.material_balance_type = 'component_phase'
    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()

    ms.fs.u.holdup._make_component_balances()

    for k in ms.fs.u.holdup.material_balance.keys():
        assert k in [(0, "phase1", "comp1"),
                     (0, "phase1", "comp2"),
                     (0, "phase2", "comp1"),
                     (0, "phase2", "comp2")]


def test_make_component_balances_rate_reactions_true(ms):
    ms.fs.u.holdup.config.has_rate_reactions = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_component_balances()

    assert hasattr(ms.fs.u.holdup, "rate_reaction_generation")
    assert hasattr(ms.fs.u.holdup, "rate_reaction_extent")
    assert hasattr(ms.fs.u.holdup, "rate_reaction_stoichiometry_constraint")


def test_make_component_balances_rate_reactions_false(ms):
    ms.fs.u.holdup.config.has_rate_reactions = False

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_component_balances()

    assert not hasattr(ms.fs.u.holdup, "rate_reaction_generation")
    assert not hasattr(ms.fs.u.holdup, "rate_reaction_extent")
    assert not hasattr(ms.fs.u.holdup,
                       "rate_reaction_stoichiometry_constraint")


def test_make_component_balances_equilibrium_reactions_true(ms):
    ms.fs.u.holdup.config.has_equilibrium_reactions = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_component_balances()

    assert hasattr(ms.fs.u.holdup, "equilibrium_reaction_generation")
    assert hasattr(ms.fs.u.holdup, "equilibrium_reaction_extent")
    assert hasattr(ms.fs.u.holdup,
                   "equilibrium_reaction_stoichiometry_constraint")


def test_make_component_balances_equilibrium_reactions_false(ms):
    ms.fs.u.holdup.config.has_equilibrium_reactions = False

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_component_balances()

    assert not hasattr(ms.fs.u.holdup, "equilibrium_reaction_generation")
    assert not hasattr(ms.fs.u.holdup, "equilibrium_reaction_extent")
    assert not hasattr(ms.fs.u.holdup,
                       "equilibrium_reaction_stoichiometry_constraint")


def test_make_component_balances_phase_equilibrium_reactions_true_total(ms):
    ms.fs.u.holdup.config.material_balance_type = 'component_total'
    ms.fs.u.holdup.config.has_phase_equilibrium = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_component_balances()

    assert not hasattr(ms.fs.u.holdup, "phase_equilibrium_generation")


def test_make_component_balances_phase_equilibrium_reactions_false_total(ms):
    ms.fs.u.holdup.config.material_balance_type = 'component_total'
    ms.fs.u.holdup.config.has_phase_equilibrium = False

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_component_balances()

    assert not hasattr(ms.fs.u.holdup, "phase_equilibrium_generation")


def test_make_component_balances_phase_equilibrium_reactions_true_phase(ms):
    ms.fs.u.holdup.config.material_balance_type = 'component_phase'
    ms.fs.u.holdup.config.has_phase_equilibrium = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_component_balances()

    assert hasattr(ms.fs.u.holdup, "phase_equilibrium_generation")


def test_make_component_balances_phase_equilibrium_reactions_false_phase(ms):
    ms.fs.u.holdup.config.material_balance_type = 'component_phase'
    ms.fs.u.holdup.config.has_phase_equilibrium = False

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_component_balances()

    assert not hasattr(ms.fs.u.holdup, "phase_equilibrium_generation")


def test_make_component_balances_mass_transfer_true(ms):
    ms.fs.u.holdup.config.has_mass_transfer = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_component_balances()

    assert hasattr(ms.fs.u.holdup, "mass_transfer_term")


def test_make_component_balances_mass_transfer_false(ms):
    ms.fs.u.holdup.config.has_mass_transfer = False

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_component_balances()

    assert not hasattr(ms.fs.u.holdup, "mass_transfer_term")


def test_make_component_balances_include_holdup_true(ms):
    ms.fs.u.holdup.config.include_holdup = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_component_balances()

    assert hasattr(ms.fs.u.holdup, "material_holdup_calculation")


def test_make_component_balances_include_holdup_false(ms):
    ms.fs.u.holdup.config.include_holdup = False

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_component_balances()

    assert not hasattr(ms.fs.u.holdup, "material_holdup_calculation")


# -----------------------------------------------------------------------------
# Testing make_element_total_balances
def test_make_element_total_balance(ms):
    ms.fs.u.holdup.config.material_balance_type = 'element_total'

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_element_total_balances()

    assert hasattr(ms.fs.u.holdup, "elemental_flow_in")
    assert hasattr(ms.fs.u.holdup, "elemental_flow_out")
    assert hasattr(ms.fs.u.holdup, "element_balance")


def test_make_element_total_balance_mass_transfer_true(ms):
    ms.fs.u.holdup.config.material_balance_type = 'element_total'
    ms.fs.u.holdup.config.has_mass_transfer = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_element_total_balances()

    assert hasattr(ms.fs.u.holdup, "elemental_mass_transfer")


def test_make_element_total_balance_mass_transfer_false(ms):
    ms.fs.u.holdup.config.material_balance_type = 'element_total'
    ms.fs.u.holdup.config.has_mass_transfer = False

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_element_total_balances()

    assert not hasattr(ms.fs.u.holdup, "elemental_mass_transfer")


def test_make_element_total_balance_include_holdup_true(ms):
    ms.fs.u.holdup.config.material_balance_type = 'element_total'
    ms.fs.u.holdup.config.include_holdup = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_element_total_balances()

    assert hasattr(ms.fs.u.holdup, "elemental_holdup_calculation")


def test_make_element_total_balance_include_holdup_false(ms):
    ms.fs.u.holdup.config.material_balance_type = 'element_total'
    ms.fs.u.holdup.config.include_holdup = False

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_element_total_balances()

    assert not hasattr(ms.fs.u.holdup, "elemental_holdup_calculation")


# -----------------------------------------------------------------------------
# Testing _make_energy_balance
def test_make_energy_balance(ms):
    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_energy_balance()

    assert hasattr(ms.fs.u.holdup, "scaling_factor_energy")
    assert hasattr(ms.fs.u.holdup, "energy_balance")


def test_make_energy_balance_has_heat_transfer_true(ms):
    ms.fs.u.holdup.config.has_heat_transfer = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_energy_balance()

    assert hasattr(ms.fs.u.holdup, "heat")


def test_make_energy_balance_has_heat_transfer_false(ms):
    ms.fs.u.holdup.config.has_heat_transfer = False

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_energy_balance()

    assert not hasattr(ms.fs.u.holdup, "heat")


def test_make_energy_balance_has_work_transfer_true(ms):
    ms.fs.u.holdup.config.has_work_transfer = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_energy_balance()

    assert hasattr(ms.fs.u.holdup, "work")


def test_make_energy_balance_has_work_transfer_false(ms):
    ms.fs.u.holdup.config.has_work_transfer = False

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_energy_balance()

    assert not hasattr(ms.fs.u.holdup, "work")


def test_make_energy_balance_include_holdup_true(ms):
    ms.fs.u.holdup.config.include_holdup = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_energy_balance()

    assert hasattr(ms.fs.u.holdup, "energy_holdup_calculation")


def test_make_energy_balance_include_holdup_false(ms):
    ms.fs.u.holdup.config.include_holdup = False

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_energy_balance()

    assert not hasattr(ms.fs.u.holdup, "energy_holdup_calculation")


# -----------------------------------------------------------------------------
# Testing _make_momentum_balance
def test_make_momentum_balance_has_pressure_change_true(ms):
    ms.fs.u.holdup.config.has_pressure_change = True

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_momentum_balance()

    assert hasattr(ms.fs.u.holdup, "deltaP")
    assert hasattr(ms.fs.u.holdup, "scaling_factor_pressure")
    assert hasattr(ms.fs.u.holdup, "momentum_balance")


def test_make_momentum_balance_has_pressure_change_false(ms):
    ms.fs.u.holdup.config.has_pressure_change = False

    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup._get_indexing_sets()
    ms.fs.u.holdup._make_vars()
    ms.fs.u.holdup._make_phase_frac()
    ms.fs.u.holdup._make_momentum_balance()

    assert not hasattr(ms.fs.u.holdup, "deltaP")
    assert hasattr(ms.fs.u.holdup, "scaling_factor_pressure")
    assert hasattr(ms.fs.u.holdup, "momentum_balance")


# -----------------------------------------------------------------------------
# Testing model_check
def test_model_check(ms):
    ms.fs.u.holdup._make_prop_blocks()
    ms.fs.u.holdup.model_check()

    assert ms.fs.u.holdup.properties_in[0].test
    assert ms.fs.u.holdup.properties_out[0].test


# -----------------------------------------------------------------------------
# Testing initialize and release_state methods
def test_initialize_and_release():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.prop = PropertyParameterBlock()

    m.fs.u = Unit()
    m.fs.u.config.declare("property_package", ConfigValue(default=m.fs.prop))
    m.fs.u.config.declare("property_package_args", ConfigValue(default={}))

    m.fs.u.holdup = Holdup0D()

    flags = m.fs.u.holdup.initialize()

    assert m.fs.u.holdup.properties_in[0].pressure.fixed
    assert m.fs.u.holdup.properties_in[0].temperature.fixed
    assert m.fs.u.holdup.properties_in[0].flow["comp1"].fixed
    assert m.fs.u.holdup.properties_in[0].flow["comp2"].fixed
    assert m.fs.u.holdup.properties_in[0].pressure.value == 10
    assert m.fs.u.holdup.properties_in[0].temperature.value == 20
    assert m.fs.u.holdup.properties_in[0].flow["comp1"].value == 30
    assert m.fs.u.holdup.properties_in[0].flow["comp2"].value == 30

    assert m.fs.u.holdup.properties_out[0].pressure.fixed
    assert m.fs.u.holdup.properties_out[0].temperature.fixed
    assert m.fs.u.holdup.properties_out[0].flow["comp1"].fixed
    assert m.fs.u.holdup.properties_out[0].flow["comp2"].fixed
    assert m.fs.u.holdup.properties_out[0].pressure.value == 10
    assert m.fs.u.holdup.properties_out[0].temperature.value == 20
    assert m.fs.u.holdup.properties_out[0].flow["comp1"].value == 30
    assert m.fs.u.holdup.properties_out[0].flow["comp2"].value == 30

    m.fs.u.holdup.release_state(flags)

    assert not m.fs.u.holdup.properties_in[0].pressure.fixed
    assert not m.fs.u.holdup.properties_in[0].temperature.fixed
    assert not m.fs.u.holdup.properties_in[0].flow["comp1"].fixed
    assert not m.fs.u.holdup.properties_in[0].flow["comp2"].fixed
    assert m.fs.u.holdup.properties_out[0].pressure.fixed
    assert m.fs.u.holdup.properties_out[0].temperature.fixed
    assert m.fs.u.holdup.properties_out[0].flow["comp1"].fixed
    assert m.fs.u.holdup.properties_out[0].flow["comp2"].fixed


def test_initialize_and_release_multiple_inlets():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.prop = PropertyParameterBlock()

    m.fs.u = Unit()
    m.fs.u.config.declare("property_package", ConfigValue(default=m.fs.prop))
    m.fs.u.config.declare("property_package_args", ConfigValue(default={}))

    m.fs.u.holdup = Holdup0D()
    m.fs.u.build_inlets(num_inlets=2)

    flags = m.fs.u.holdup.initialize()

    assert m.fs.u.holdup.properties_in[0].pressure.fixed
    assert m.fs.u.holdup.properties_in[0].temperature.fixed
    assert m.fs.u.holdup.properties_in[0].flow["comp1"].fixed
    assert m.fs.u.holdup.properties_in[0].flow["comp2"].fixed
    assert m.fs.u.holdup.properties_out[0].pressure.fixed
    assert m.fs.u.holdup.properties_out[0].temperature.fixed
    assert m.fs.u.holdup.properties_out[0].flow["comp1"].fixed
    assert m.fs.u.holdup.properties_out[0].flow["comp2"].fixed

    assert m.fs.u.holdup.inlet_mixer.properties[0, "1"].pressure.fixed
    assert m.fs.u.holdup.inlet_mixer.properties[0, "1"].temperature.fixed
    assert m.fs.u.holdup.inlet_mixer.properties[0, "1"].flow["comp1"].fixed
    assert m.fs.u.holdup.inlet_mixer.properties[0, "1"].flow["comp2"].fixed
    assert m.fs.u.holdup.inlet_mixer.properties[0, "1"].pressure.value == 10
    assert m.fs.u.holdup.inlet_mixer.properties[0, "1"].temperature.value == 20
    assert (m.fs.u.holdup.inlet_mixer.properties[0, "1"].flow["comp1"].value ==
            30)
    assert (m.fs.u.holdup.inlet_mixer.properties[0, "1"].flow["comp2"].value ==
            30)

    m.fs.u.holdup.release_state(flags)

    assert m.fs.u.holdup.properties_in[0].pressure.fixed
    assert m.fs.u.holdup.properties_in[0].temperature.fixed
    assert m.fs.u.holdup.properties_in[0].flow["comp1"].fixed
    assert m.fs.u.holdup.properties_in[0].flow["comp2"].fixed
    assert m.fs.u.holdup.properties_out[0].pressure.fixed
    assert m.fs.u.holdup.properties_out[0].temperature.fixed
    assert m.fs.u.holdup.properties_out[0].flow["comp1"].fixed
    assert m.fs.u.holdup.properties_out[0].flow["comp2"].fixed

    assert not m.fs.u.holdup.inlet_mixer.properties[0, "1"].pressure.fixed
    assert not m.fs.u.holdup.inlet_mixer.properties[0, "1"].temperature.fixed
    assert not m.fs.u.holdup.inlet_mixer.properties[0, "1"].flow["comp1"].fixed
    assert not m.fs.u.holdup.inlet_mixer.properties[0, "1"].flow["comp2"].fixed
    assert not m.fs.u.holdup.inlet_mixer.properties[0, "2"].pressure.fixed
    assert not m.fs.u.holdup.inlet_mixer.properties[0, "2"].temperature.fixed
    assert not m.fs.u.holdup.inlet_mixer.properties[0, "2"].flow["comp1"].fixed
    assert not m.fs.u.holdup.inlet_mixer.properties[0, "2"].flow["comp2"].fixed
