##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
import idaes.core.util.convergence.convergence_base as cb
import idaes.models.convergence.pressure_changer.pressure_changer_simulation as pcs
import pyomo.environ as pe
'''
This module contains the code for convergence testing of the PressureChanger model
'''

class PressureChangerConvergenceEvaluation(cb.ConvergenceEvaluation):
    def get_specification(self):
        """
        Returns the convergence evaluation specification for the PressureChanger unit model

        Returns
        -------
           ConvergenceEvaluationSpecification
        """
        s = cb.ConvergenceEvaluationSpecification()

        s.add_sampled_input(name='Inlet_Flowrate', pyomo_path='fs_obj.C01.holdup.properties_in[0].flow_mol_comp["H2O"]',
                  lower= 1, upper=1e6, mean=5e5, std=2e5)

        s.add_sampled_input(name='Inlet_Temperature', pyomo_path='fs_obj.C01.holdup.properties_in[0].temperature',
                  lower= 200, upper=1000, mean=600, std=200)
        return s

    def get_initialized_model(self):
        """
        Returns an initialized model for the PressureChanger unit model convergence evaluation

        Returns
        -------
           Pyomo model : returns a pyomo model of the PressureChanger unit
        """
        # Create a Concrete Model as the top level object
        m = pcs.build_model()

        # Fix the inputs
        pcs.setInputs(m)

        # Create a solver for initialization
        stee = True
        opt = self.get_solver()

        m.fs_obj.model_check()

        # Initialize the model
        pcs.initialize_model(m, opt, stee)

        # return the initialized model
        return m

    def get_solver(self):
        """
        Returns an ipopt solver object with the desired options for convergence evaluation (and initialization)

        Returns
        -------
           Pyomo solver
        """
        opt = pe.SolverFactory('ipopt')
        opt.options = {'tol': 1e-6,
                       'mu_init': 1e-8,
                       'bound_push': 1e-8}
        return opt

