##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Standard IDAES CSTR model.
"""
from __future__ import division

# Import Python libraries
import logging

# Import Pyomo libraries
from pyomo.environ import Var
from pyomo.common.config import ConfigValue

# Import IDAES cores
from idaes.core import UnitBlockData, declare_process_block_class, \
                        Holdup0D, CONFIG_Base
from idaes.core.util.config import is_parameter_block, list_of_strings
from idaes.core.util.misc import add_object_ref

__author__ = "Andrew Lee"


# Set up logger
logger = logging.getLogger('idaes.unit_model')


@declare_process_block_class("CSTR")
class CSTRData(UnitBlockData):
    """
    Standard CSTR Unit Model Class
    """
    CONFIG = CONFIG_Base()
    # Set default values of inherited attributes
    CONFIG.get('include_holdup')._default = False
    CONFIG.get("material_balance_type")._default = 'component_phase'
    CONFIG.get("energy_balance_type")._default = 'total'
    CONFIG.get("momentum_balance_type")._default = 'total'
    CONFIG.get("has_rate_reactions")._default = True
    CONFIG.get("has_equilibrium_reactions")._default = True
    CONFIG.get("has_phase_equilibrium")._default = False
    CONFIG.get("has_mass_transfer")._default = False
    CONFIG.get("has_heat_transfer")._default = True
    CONFIG.get("has_work_transfer")._default = False
    CONFIG.get("has_pressure_change")._default = False
    # Add unit model attributes
    CONFIG.declare("property_package", ConfigValue(
        default=None,
        domain=is_parameter_block,
        description="Property package to use for holdup",
        doc="""Property parameter object used to define property calculations
            (default = 'use_parent_value')
                - 'use_parent_value' - get package from parent (default = None)
                - a ParameterBlock object"""))
    CONFIG.declare("property_package_args", ConfigValue(
        default={},
        description="Arguments to use for constructing property packages",
        doc="""A dict of arguments to be passed to the PropertyBlockData
            and used when constructing these
            (default = 'use_parent_value')
                - 'use_parent_value' - get package from parent (default = None)
                - a dict (see property package for documentation)"""))
    CONFIG.declare("inlet_list", ConfigValue(
        domain=list_of_strings,
        description="List of inlet names",
        doc="""A list containing names of inlets (default = None)
                - None - default single inlet
                - list - a list of names for inlets"""))
    CONFIG.declare("num_inlets", ConfigValue(
        domain=int,
        description="Number of inlets to unit",
        doc="""Argument indication number (int) of inlets to construct
            (default = None). Not used if inlet_list arg is provided.
                - None - use inlet_list arg instead
                - int - Inlets will be named with sequential numbers from 1
                        to num_inlets"""))
    CONFIG.declare("outlet_list", ConfigValue(
        domain=list_of_strings,
        description="List of outlet names",
        doc="""A list containing names of outlets (default = None)
                - None - default single outlet
                - list - a list of names for outlets"""))
    CONFIG.declare("num_outlets", ConfigValue(
        domain=int,
        description="Number of outlets to unit",
        doc="""Argument indication number (int) of outlets to construct
            (default = None). Not used if outlet_list arg is provided.
                - None - use outlet_list arg instead
                - int - Outlets will be named with sequential numbers from 1
                        to num_outlets"""))

    def build(self):
        """
        Begin building model (pre-DAE transformation).

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(CSTRData, self).build()

        # Build Holdup Block
        self.holdup = Holdup0D()

        # Set Unit Geometry and holdup Volume
        self._set_geometry()

        # Construct performance equations
        self._make_performance()

    def post_transform_build(self):
        """
        Continue model construction after DAE transformation.

        Args:
            None

        Returns:
            None
        """
        # Construct Inlets
        self.build_inlets(inlets=self.config.inlet_list,
                          num_inlets=self.config.num_inlets)

        # Build Outlets
        self.build_outlets(outlets=self.config.outlet_list,
                           num_outlets=self.config.num_outlets)

    def _set_geometry(self):
        """
        Define the geometry of the unit as necessary, and link to holdup volume

        Args:
            None

        Returns:
            None
        """
        # For this case, just create a reference to holdup volume
        if self.config.include_holdup is True:
            add_object_ref(self, "volume", self.holdup.volume)
        else:
            # Add a volume variable, as CSTRs always need a volume
            self.volume = Var(self.time,
                              initialize=1.0,
                              doc="Reactor volume")

    def _make_performance(self):
        """
        Define constraints which describe the behaviour of the unit model.

        Args:
            None

        Returns:
            None
        """
        # Add constraint relating extents to rate of reaction
        if hasattr(self.holdup, "rate_reaction_idx"):
            add_object_ref(self, "rate_reaction_idx",
                           self.holdup.rate_reaction_idx)

            @self.Constraint(self.time,
                             self.rate_reaction_idx,
                             doc="Extents of reaction")
            def rate_reaction_extents(b, t, k):
                return b.holdup.rate_reaction_extent[t, k] == (
                            b.volume[t] *
                            b.holdup.properties_out[t].reaction_rate[k])

        # Set references to balance terms at unit level
        if (self.config.has_heat_transfer is True and
                self.config.energy_balance_type != 'none'):
            add_object_ref(self, "heat", self.holdup.heat)
