##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Tests for Translator unit model.

Author: Andrew Lee <andrew.lee@netl.doe.gov>
"""
import pytest
from pyomo.environ import ConcreteModel
from idaes.core import FlowsheetBlockData, declare_process_block_class
from idaes.models.translator import Translator, TranslatorData

# Import property package for testing
from os.path import abspath, dirname, join
from pyutilib.misc import import_file
example1 = join(dirname(abspath(__file__)), '..', '..', '..', 'examples',
                'core', 'examples', 'reaction_properties_demo.py')
p1 = import_file(example1)
example2 = join(dirname(abspath(__file__)), '..', '..', '..', 'examples',
                'core', 'examples', 'steam_properties_demo.py')
p2 = import_file(example2)


# -----------------------------------------------------------------------------
# General test classes
@declare_process_block_class("Flowsheet")
class _Flowsheet(FlowsheetBlockData):
    def build(self):
        super(_Flowsheet, self).build()


@declare_process_block_class("TranslatorFrame")
class _Translator(TranslatorData):
    def build(self):
        pass


@pytest.fixture()
def m():
    """
    Build a simple flowsheet model for build tests from.
    """
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props1 = p1.PropertyParameterBlock()
    m.fs.props2 = p2.PropertyParameterBlock()
    m.fs.trans = TranslatorFrame(inlet_property_package=m.fs.props1,
                                 outlet_property_package=m.fs.props2)
    return m


# -----------------------------------------------------------------------------
# Test class methods
def test_default_config_block(m):
    assert m.fs.trans.config.dynamic == 'use_parent_value'
    assert m.fs.trans.config.inlet_property_package == m.fs.props1
    assert m.fs.trans.config.inlet_property_package_args == {}
    assert m.fs.trans.config.outlet_property_package == m.fs.props2
    assert m.fs.trans.config.outlet_property_package_args == {}

    assert not m.fs.trans.config.has_equilibrium_reactions
    m.fs.trans.config.has_equilibrium_reactions = True
    with pytest.raises(ValueError):
        m.fs.trans.config.has_equilibrium_reactions = "foo"
    with pytest.raises(ValueError):
        m.fs.trans.config.has_equilibrium_reactions = 10

    assert not m.fs.trans.config.has_phase_equilibrium
    m.fs.trans.config.has_phase_equilibrium = True
    with pytest.raises(ValueError):
        m.fs.trans.config.has_phase_equilibrium = "foo"
    with pytest.raises(ValueError):
        m.fs.trans.config.has_phase_equilibrium = 10


def test_super_build(m):
    super(TranslatorData, m.fs.trans).build()
    assert not m.fs.trans.config.dynamic


def test_create_property_blocks(m):
    m.fs.trans.config.has_equilibrium_reactions = True
    m.fs.trans.config.has_phase_equilibrium = True

    super(TranslatorData, m.fs.trans).build()
    m.fs.trans._create_property_blocks()

    assert hasattr(m.fs.trans, "inlet_properties")
    for k in m.fs.trans.inlet_properties[0].component_list:
        assert k in ["a", "b", "c", "d", "e", "f"]

    assert hasattr(m.fs.trans, "outlet_properties")
    assert m.fs.trans.outlet_properties[0].component_list == ["H2O"]

    assert m.fs.trans.inlet_properties[0].config.has_sum_fractions is False
    assert (m.fs.trans.inlet_properties[0].config.
            calculate_equilibrium_reactions is True)
    assert (m.fs.trans.inlet_properties[0].config.calculate_phase_equilibrium
            is True)
    assert m.fs.trans.outlet_properties[0].config.has_sum_fractions is True
    assert (m.fs.trans.outlet_properties[0].config.
            calculate_equilibrium_reactions is True)
    assert (m.fs.trans.outlet_properties[0].config.calculate_phase_equilibrium
            is True)


def test_create_property_blocks_get_default():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props1 = p1.PropertyParameterBlock()
    m.fs.props2 = p2.PropertyParameterBlock()
    m.fs.config.default_property_package = m.fs.props1
    m.fs.trans = TranslatorFrame(outlet_property_package=m.fs.props2)

    super(TranslatorData, m.fs.trans).build()
    m.fs.trans._create_property_blocks()

    assert hasattr(m.fs.trans, "inlet_properties")
    for k in m.fs.trans.inlet_properties[0].component_list:
        assert k in ["a", "b", "c", "d", "e", "f"]

    assert hasattr(m.fs.trans, "outlet_properties")
    assert m.fs.trans.outlet_properties[0].component_list == ["H2O"]


def test_create_property_blocks_get_default_none():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props1 = p1.PropertyParameterBlock()
    m.fs.props2 = p2.PropertyParameterBlock()
    m.fs.trans = TranslatorFrame(outlet_property_package=m.fs.props2)

    super(TranslatorData, m.fs.trans).build()
    with pytest.raises(ValueError):
        m.fs.trans._create_property_blocks()


# -----------------------------------------------------------------------------
# Test post_transform_build and connections
def test_post_transform_build():
    m = ConcreteModel()
    m.fs = Flowsheet(dynamic=False)
    m.fs.props1 = p1.PropertyParameterBlock()
    m.fs.props2 = p2.PropertyParameterBlock()
    m.fs.trans = Translator(inlet_property_package=m.fs.props1,
                            outlet_property_package=m.fs.props2)

    m.fs.trans.post_transform_build()

    assert hasattr(m.fs.trans, "inlet")
    for k in m.fs.trans.inlet[0].vars:
        assert k in ["flow_mol_comp", "enth_mol", "pressure"]
    assert hasattr(m.fs.trans, "outlet")
    for k in m.fs.trans.outlet[0].vars:
        assert k in ["flow_mol_comp", "temperature", "pressure"]
