##############################################################################
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
# 
# Please see the files COPYRIGHT.txt and LICENSE.txt for full copyright and
# license information, respectively. Both files are also available online
# at the URL "https://github.com/IDAES/idaes".
##############################################################################
"""
Some very basic tests for the 'contrib' directories
"""
from __future__ import print_function

import importlib
import os
import sys
import threading
import time
#
import pytest
#
import idaes

TIMEOUT_SEC = 3  # any longer means it's running something


@pytest.mark.skip()
def test_syntax():
    """Make sure all the 'contrib' modules at least are
       valid syntax.
    """
    _test_modules(SyntaxError)


# Thread communication through an array
g_results = []


def _test_modules(exc_class):
    global g_results
    print('Test import of contrib modules for exception type "{}"'
          .format(str(exc_class)))
    failed = []
    for i, m in enumerate(find_contrib_modules()):
        g_results.append(None)  # make a spot to report results
        failure = can_import_module(m, exc_class, i)
        if failure:
            failed.append((m, failure))
    if failed:
        msg = 'Modules raising "{}" on import:\n'.format(str(exc_class))
        maxlen = max([len(f[0]) for f in failed])
        row = '  {{:{:d}s}} {{}}\n'.format(maxlen)
        for name, reason in failed:
            msg += row.format(name, reason)
    else:
        msg = ''
    num_failed = len(failed)
    assert num_failed == 0, msg


def find_contrib_modules():
    """Find modules in 'contrib' and return them (generator).
    """
    root = os.path.dirname(idaes.__file__)
    root_prefix = len(root) + 1  # +1 is for '/' between prefix and rest
    for dirname, _, filenames in os.walk(os.path.join(root, 'contrib')):
        if '__init__.py' in filenames:  # this is a package
            dirname_package = dirname[root_prefix:].replace('/', '.')
            package = 'idaes.' + dirname_package
            for fname in filenames:
                if fname.endswith('.py'):
                    if fname == '__init__.py':
                        module = package
                    else:
                        module = package + '.' + fname[:-3]
                    yield module


def can_import_module(m, exc_class, i):
    """Called serially to import a module.
    """
    print('Attempting to import module: {} .. '.format(m), end='')
    sys.stdout.flush()
    result = None
    thr = threading.Thread(target=try_import, args=(m, exc_class, i))
    thr.setDaemon(True)
    thr.start()
    thr.join(TIMEOUT_SEC)
    if thr.is_alive():
        print('TIMEOUT')
        # don't set result: this is not a failure
    else:
        # retrieve the result
        result = g_results[i]
    return result


def try_import(m, exc_class, i):
    """Called in a thread to import a module.

    Print status *unless* we ran too long, in which case the
    caller will have already printed that we timed out.

    Args:
        m (str):  Name of module
        exc_class (Exception): Target exception class
        i (int): This thread's index into result array
    """
    result, t0 = None, time.time()
    try:
        importlib.import_module(m)
        dt = time.time() - t0
        if dt < TIMEOUT_SEC:
            print('SUCCESS')
    except exc_class as err:
        dt = time.time() - t0
        result = err
        if dt < TIMEOUT_SEC:
            print('FAILED : {}\n'.format(err))
    except Exception as err:
        dt = time.time() - t0
        if dt < TIMEOUT_SEC:
            errclass = err.__class__.__name__
            print('IGNORED: {}'.format(errclass))
    # Report result by setting value in global array
    g_results[i] = result

